@@ -63,6 +63,7 @@
 
 #include <errno.h>
 #include <verify.h>
+#include <alignof.h>
 #include "libguile/__scm.h"
 
 /* Include headers for those files central to the implementation.  The
@@ -117,9 +118,13 @@
 #define max(A, B) ((A) >= (B) ? (A) : (B))
 #endif
 
+/* Return the first integer greater than or equal to LEN such that
+   LEN % ALIGN == 0.  Return LEN if ALIGN is zero.  */
+#define ROUND_UP(len, align)					\
+  ((align) ? (((len) - 1UL) | ((align) - 1UL)) + 1UL : (len))
 
 
-#if GUILE_USE_64_CALLS && HAVE_STAT64
+#if GUILE_USE_64_CALLS && defined(HAVE_STAT64)
 #define CHOOSE_LARGEFILE(foo,foo64)     foo64
 #else
 #define CHOOSE_LARGEFILE(foo,foo64)     foo
@@ -178,7 +183,7 @@
 
 /* Major and minor versions must be single characters. */
 #define SCM_OBJCODE_MAJOR_VERSION 0
-#define SCM_OBJCODE_MINOR_VERSION P
+#define SCM_OBJCODE_MINOR_VERSION R
 #define SCM_OBJCODE_MAJOR_VERSION_STRING        \
   SCM_CPP_STRINGIFY(SCM_OBJCODE_MAJOR_VERSION)
 #define SCM_OBJCODE_MINOR_VERSION_STRING        \
@@ -192,6 +197,35 @@
 #define SCM_OBJCODE_COOKIE                              \
   "GOOF-" SCM_OBJCODE_MACHINE_VERSION_STRING "---"
 
+
+/* Cells have to be 8-byte aligned.  Use `scm_t_aligned_cell' when not
+   allocating on the heap to have this guarantee.  This is similar to the
+   `SCM_ALIGNED' macro but provides an option likely to work with compilers
+   other than GCC.  */
+
+#ifdef __GNUC__
+
+struct scm_aligned_cell
+{
+  scm_t_cell cell __attribute__ ((__aligned__ (8)));
+};
+
+typedef struct scm_aligned_cell scm_t_aligned_cell;
+
+#else /* !__GNUC__ */
+
+union scm_aligned_cell
+{
+  double alignment;
+  scm_t_cell cell;
+};
+
+typedef union scm_aligned_cell scm_t_aligned_cell;
+
+#endif /* !__GNUC__ */
+
+/* Make sure we get the right alignment.  */
+verify (alignof (scm_t_aligned_cell) >= 8);
 
 #endif  /* SCM__SCM_H */
 
