@@ -24,7 +24,6 @@
 #endif
 
 #include <alloca.h>
-#include <assert.h>
 
 #include "libguile/__scm.h"
 
@@ -37,6 +36,7 @@
 #include "libguile/deprecation.h"
 #include "libguile/dynwind.h"
 #include "libguile/eq.h"
+#include "libguile/expand.h"
 #include "libguile/feature.h"
 #include "libguile/fluids.h"
 #include "libguile/goops.h"
@@ -106,10 +106,40 @@ static scm_t_bits scm_tc16_boot_closure;
 #define BOOT_CLOSURE_P(obj) SCM_TYP16_PREDICATE (scm_tc16_boot_closure, (obj))
 #define BOOT_CLOSURE_CODE(x) SCM_SMOB_OBJECT (x)
 #define BOOT_CLOSURE_ENV(x) SCM_SMOB_OBJECT_2 (x)
-#define BOOT_CLOSURE_NUM_REQUIRED_ARGS(x) SCM_I_INUM (CAR (BOOT_CLOSURE_CODE (x)))
-#define BOOT_CLOSURE_HAS_REST_ARGS(x) scm_is_true (CADR (BOOT_CLOSURE_CODE (x)))
-#define BOOT_CLOSURE_BODY(x) CDDR (BOOT_CLOSURE_CODE (x))
-
+#define BOOT_CLOSURE_BODY(x) CAR (BOOT_CLOSURE_CODE (x))
+#define BOOT_CLOSURE_NUM_REQUIRED_ARGS(x) SCM_I_INUM (CADR (BOOT_CLOSURE_CODE (x)))
+#define BOOT_CLOSURE_IS_FIXED(x) scm_is_null (CDDR (BOOT_CLOSURE_CODE (x)))
+/* NB: One may only call the following accessors if the closure is not FIXED. */
+#define BOOT_CLOSURE_HAS_REST_ARGS(x) scm_is_true (CADDR (BOOT_CLOSURE_CODE (x)))
+#define BOOT_CLOSURE_IS_REST(x) scm_is_null (CDDDR (BOOT_CLOSURE_CODE (x)))
+/* NB: One may only call the following accessors if the closure is not REST. */
+#define BOOT_CLOSURE_IS_FULL(x) (1)
+#define BOOT_CLOSURE_PARSE_FULL(fu_,body,nargs,rest,nopt,kw,inits,alt)    \
+  do { SCM fu = fu_;                                            \
+    body = CAR (fu); fu = CDR (fu);                             \
+                                                                \
+    rest = kw = alt = SCM_BOOL_F;                               \
+    inits = SCM_EOL;                                            \
+    nopt = 0;                                                   \
+                                                                \
+    nreq = SCM_I_INUM (CAR (fu)); fu = CDR (fu);                \
+    if (scm_is_pair (fu))                                       \
+      {                                                         \
+        rest = CAR (fu); fu = CDR (fu);                         \
+        if (scm_is_pair (fu))                                   \
+          {                                                     \
+            nopt = SCM_I_INUM (CAR (fu)); fu = CDR (fu);        \
+            kw = CAR (fu); fu = CDR (fu);                       \
+            inits = CAR (fu); fu = CDR (fu);                    \
+            alt = CAR (fu);                                     \
+          }                                                     \
+      }                                                         \
+  } while (0)
+static void prepare_boot_closure_env_for_apply (SCM proc, SCM args,
+                                                SCM *out_body, SCM *out_env);
+static void prepare_boot_closure_env_for_eval (SCM proc, unsigned int argc,
+                                               SCM exps, SCM *out_body,
+                                               SCM *inout_env);
 
 
 #define CAR(x)   SCM_CAR(x)
@@ -130,21 +160,21 @@ static void error_used_before_defined (void)
              "Variable used before given a value", SCM_EOL, SCM_BOOL_F);
 }
 
-int
-scm_badargsp (SCM formals, SCM args)
+static void error_invalid_keyword (SCM proc)
 {
-  while (!scm_is_null (formals))
-    {
-      if (!scm_is_pair (formals)) 
-        return 0;
-      if (scm_is_null (args)) 
-        return 1;
-      formals = CDR (formals);
-      args = CDR (args);
-    }
-  return !scm_is_null (args) ? 1 : 0;
+  scm_error_scm (scm_from_locale_symbol ("keyword-argument-error"), proc,
+                 scm_from_locale_string ("Invalid keyword"), SCM_EOL,
+                 SCM_BOOL_F);
 }
 
+static void error_unrecognized_keyword (SCM proc)
+{
+  scm_error_scm (scm_from_locale_symbol ("keyword-argument-error"), proc,
+                 scm_from_locale_string ("Unrecognized keyword"), SCM_EOL,
+                 SCM_BOOL_F);
+}
+
+
 /* the environment:
    (VAL ... . MOD)
    If MOD is #f, it means the environment was captured before modules were
@@ -253,25 +283,7 @@ eval (SCM x, SCM env)
        * ARGS is the list of arguments. */
       if (BOOT_CLOSURE_P (proc))
         {
-          int nreq = BOOT_CLOSURE_NUM_REQUIRED_ARGS (proc);
-          SCM new_env = BOOT_CLOSURE_ENV (proc);
-          if (BOOT_CLOSURE_HAS_REST_ARGS (proc))
-            {
-              if (SCM_UNLIKELY (scm_ilength (args) < nreq))
-                scm_wrong_num_args (proc);
-              for (; nreq; nreq--, args = CDR (args))
-                new_env = scm_cons (CAR (args), new_env);
-              new_env = scm_cons (args, new_env);
-            }
-          else
-            {
-              if (SCM_UNLIKELY (scm_ilength (args) != nreq))
-                scm_wrong_num_args (proc);
-              for (; scm_is_pair (args); args = CDR (args))
-                new_env = scm_cons (CAR (args), new_env);
-            }
-          x = BOOT_CLOSURE_BODY (proc);
-          env = new_env;
+          prepare_boot_closure_env_for_apply (proc, args, &x, &env);
           goto loop;
         }
       else
@@ -285,31 +297,7 @@ eval (SCM x, SCM env)
 
       if (BOOT_CLOSURE_P (proc))
         {
-          int nreq = BOOT_CLOSURE_NUM_REQUIRED_ARGS (proc);
-          SCM new_env = BOOT_CLOSURE_ENV (proc);
-          if (BOOT_CLOSURE_HAS_REST_ARGS (proc))
-            {
-              if (SCM_UNLIKELY (argc < nreq))
-                scm_wrong_num_args (proc);
-              for (; nreq; nreq--, mx = CDR (mx))
-                new_env = scm_cons (eval (CAR (mx), env), new_env);
-              {
-                SCM rest = SCM_EOL;
-                for (; scm_is_pair (mx); mx = CDR (mx))
-                  rest = scm_cons (eval (CAR (mx), env), rest);
-                new_env = scm_cons (scm_reverse (rest),
-                                    new_env);
-              }
-            }
-          else
-            {
-              for (; scm_is_pair (mx); mx = CDR (mx), nreq--)
-                new_env = scm_cons (eval (CAR (mx), env), new_env);
-              if (SCM_UNLIKELY (nreq != 0))
-                scm_wrong_num_args (proc);
-            }
-          x = BOOT_CLOSURE_BODY (proc);
-          env = new_env;
+          prepare_boot_closure_env_for_eval (proc, argc, mx, &x, &env);
           goto loop;
         }
       else
@@ -844,11 +832,9 @@ scm_for_each (SCM proc, SCM arg1, SCM args)
 static SCM
 scm_c_primitive_eval (SCM exp)
 {
-  SCM transformer = scm_current_module_transformer ();
-  if (scm_is_true (transformer))
-    exp = scm_call_1 (transformer, exp);
-  exp = scm_memoize_expression (exp);
-  return eval (exp, SCM_EOL);
+  if (!SCM_EXPANDED_P (exp))
+    exp = scm_call_1 (scm_current_module_transformer (), exp);
+  return eval (scm_memoize_expression (exp), SCM_EOL);
 }
 
 static SCM var_primitive_eval;
@@ -921,28 +907,210 @@ scm_apply (SCM proc, SCM arg1, SCM args)
   return scm_vm_apply (scm_the_vm (), proc, args);
 }
 
+static void
+prepare_boot_closure_env_for_apply (SCM proc, SCM args,
+                                    SCM *out_body, SCM *out_env)
+{
+  int nreq = BOOT_CLOSURE_NUM_REQUIRED_ARGS (proc);
+  SCM env = BOOT_CLOSURE_ENV (proc);
+  
+  if (BOOT_CLOSURE_IS_FIXED (proc)
+      || (BOOT_CLOSURE_IS_REST (proc)
+          && !BOOT_CLOSURE_HAS_REST_ARGS (proc)))
+    {
+      if (SCM_UNLIKELY (scm_ilength (args) != nreq))
+        scm_wrong_num_args (proc);
+      for (; scm_is_pair (args); args = CDR (args))
+        env = scm_cons (CAR (args), env);
+      *out_body = BOOT_CLOSURE_BODY (proc);
+      *out_env = env;
+    }
+  else if (BOOT_CLOSURE_IS_REST (proc))
+    {
+      if (SCM_UNLIKELY (scm_ilength (args) < nreq))
+        scm_wrong_num_args (proc);
+      for (; nreq; nreq--, args = CDR (args))
+        env = scm_cons (CAR (args), env);
+      env = scm_cons (args, env);
+      *out_body = BOOT_CLOSURE_BODY (proc);
+      *out_env = env;
+    }
+  else
+    {
+      int i, argc, nreq, nopt;
+      SCM body, rest, kw, inits, alt;
+      SCM mx = BOOT_CLOSURE_CODE (proc);
+      
+    loop:
+      BOOT_CLOSURE_PARSE_FULL (mx, body, nargs, rest, nopt, kw, inits, alt);
+
+      argc = scm_ilength (args);
+      if (argc < nreq)
+        {
+          if (scm_is_true (alt))
+            {
+              mx = alt;
+              goto loop;
+            }
+          else
+            scm_wrong_num_args (proc);
+        }
+      if (scm_is_false (kw) && argc > nreq + nopt && scm_is_false (rest))
+        {
+          if (scm_is_true (alt))
+            {
+              mx = alt;
+              goto loop;
+            }
+          else
+            scm_wrong_num_args (proc);
+        }
+
+      for (i = 0; i < nreq; i++, args = CDR (args))
+        env = scm_cons (CAR (args), env);
+
+      if (scm_is_false (kw))
+        {
+          /* Optional args (possibly), but no keyword args. */
+          for (; i < argc && i < nreq + nopt;
+               i++, args = CDR (args))
+            {
+              env = scm_cons (CAR (args), env);
+              inits = CDR (inits);
+            }
+              
+          for (; i < nreq + nopt; i++, inits = CDR (inits))
+            env = scm_cons (eval (CAR (inits), env), env);
+
+          if (scm_is_true (rest))
+            env = scm_cons (args, env);
+        }
+      else
+        {
+          SCM aok;
+
+          aok = CAR (kw);
+          kw = CDR (kw);
+
+          /* Keyword args. As before, but stop at the first keyword. */
+          for (; i < argc && i < nreq + nopt && !scm_is_keyword (CAR (args));
+               i++, args = CDR (args), inits = CDR (inits))
+            env = scm_cons (CAR (args), env);
+              
+          for (; i < nreq + nopt; i++, inits = CDR (inits))
+            env = scm_cons (eval (CAR (inits), env), env);
+
+          if (scm_is_true (rest))
+            {
+              env = scm_cons (args, env);
+              i++;
+            }
+
+          /* Now fill in env with unbound values, limn the rest of the args for
+             keywords, and fill in unbound values with their inits. */
+          {
+            int imax = i - 1;
+            int kw_start_idx = i;
+            SCM walk, k, v;
+            for (walk = kw; scm_is_pair (walk); walk = CDR (walk))
+              if (SCM_I_INUM (CDAR (walk)) > imax)
+                imax = SCM_I_INUM (CDAR (walk));
+            for (; i <= imax; i++)
+              env = scm_cons (SCM_UNDEFINED, env);
+
+            if (scm_is_pair (args) && scm_is_pair (CDR (args)))
+              for (; scm_is_pair (args) && scm_is_pair (CDR (args));
+                   args = CDR (args))
+                {
+                  k = CAR (args); v = CADR (args);
+                  if (!scm_is_keyword (k))
+                    {
+                      if (scm_is_true (rest))
+                        continue;
+                      else
+                        break;
+                    }
+                  for (walk = kw; scm_is_pair (walk); walk = CDR (walk))
+                    if (scm_is_eq (k, CAAR (walk)))
+                      {
+                        /* Well... ok, list-set! isn't the nicest interface, but
+                           hey. */
+                        int iset = imax - SCM_I_INUM (CDAR (walk));
+                        scm_list_set_x (env, SCM_I_MAKINUM (iset), v);
+                        args = CDR (args);
+                        break;
+                      }
+                  if (scm_is_null (walk) && scm_is_false (aok))
+                    error_unrecognized_keyword (proc);
+                }
+            if (scm_is_pair (args) && scm_is_false (rest))
+              error_invalid_keyword (proc);
+
+            /* Now fill in unbound values, evaluating init expressions in their
+               appropriate environment. */
+            for (i = imax - kw_start_idx; scm_is_pair (inits); i--, inits = CDR (inits))
+              {
+                SCM tail = scm_list_tail (env, SCM_I_MAKINUM (i));
+                if (SCM_UNBNDP (CAR (tail)))
+                  SCM_SETCAR (tail, eval (CAR (inits), CDR (tail)));
+              }
+          }
+        }
+
+      *out_body = body;
+      *out_env = env;
+    }
+}
+
+static void
+prepare_boot_closure_env_for_eval (SCM proc, unsigned int argc,
+                                   SCM exps, SCM *out_body, SCM *inout_env)
+{
+  int nreq = BOOT_CLOSURE_NUM_REQUIRED_ARGS (proc);
+  SCM new_env = BOOT_CLOSURE_ENV (proc);
+  if (BOOT_CLOSURE_IS_FIXED (proc)
+      || (BOOT_CLOSURE_IS_REST (proc)
+          && !BOOT_CLOSURE_HAS_REST_ARGS (proc)))
+    {
+      for (; scm_is_pair (exps); exps = CDR (exps), nreq--)
+        new_env = scm_cons (eval (CAR (exps), *inout_env), new_env);
+      if (SCM_UNLIKELY (nreq != 0))
+        scm_wrong_num_args (proc);
+      *out_body = BOOT_CLOSURE_BODY (proc);
+      *inout_env = new_env;
+    }
+  else if (BOOT_CLOSURE_IS_REST (proc))
+    {
+      if (SCM_UNLIKELY (argc < nreq))
+        scm_wrong_num_args (proc);
+      for (; nreq; nreq--, exps = CDR (exps))
+        new_env = scm_cons (eval (CAR (exps), *inout_env), new_env);
+      {
+        SCM rest = SCM_EOL;
+        for (; scm_is_pair (exps); exps = CDR (exps))
+          rest = scm_cons (eval (CAR (exps), *inout_env), rest);
+        new_env = scm_cons (scm_reverse (rest),
+                            new_env);
+      }
+      *out_body = BOOT_CLOSURE_BODY (proc);
+      *inout_env = new_env;
+    }
+  else
+    {
+      SCM args = SCM_EOL;
+      for (; scm_is_pair (exps); exps = CDR (exps))
+        args = scm_cons (eval (CAR (exps), *inout_env), args);
+      args = scm_reverse_x (args, SCM_UNDEFINED);
+      prepare_boot_closure_env_for_apply (proc, args, out_body, inout_env);
+    }
+}
 
 static SCM
 boot_closure_apply (SCM closure, SCM args)
 {
-  int nreq = BOOT_CLOSURE_NUM_REQUIRED_ARGS (closure);
-  SCM new_env = BOOT_CLOSURE_ENV (closure);
-  if (BOOT_CLOSURE_HAS_REST_ARGS (closure))
-    {
-      if (SCM_UNLIKELY (scm_ilength (args) < nreq))
-        scm_wrong_num_args (closure);
-      for (; nreq; nreq--, args = CDR (args))
-        new_env = scm_cons (CAR (args), new_env);
-      new_env = scm_cons (args, new_env);
-    }
-  else
-    {
-      if (SCM_UNLIKELY (scm_ilength (args) != nreq))
-        scm_wrong_num_args (closure);
-      for (; scm_is_pair (args); args = CDR (args))
-        new_env = scm_cons (CAR (args), new_env);
-    }
-  return eval (BOOT_CLOSURE_BODY (closure), new_env);
+  SCM body, env;
+  prepare_boot_closure_env_for_apply (closure, args, &body, &env);
+  return eval (body, env);
 }
 
 static int
@@ -954,8 +1122,9 @@ boot_closure_print (SCM closure, SCM port, scm_print_state *pstate)
   scm_putc (' ', port);
   args = scm_make_list (scm_from_int (BOOT_CLOSURE_NUM_REQUIRED_ARGS (closure)),
                         scm_from_locale_symbol ("_"));
-  if (BOOT_CLOSURE_HAS_REST_ARGS (closure))
+  if (!BOOT_CLOSURE_IS_FIXED (closure) && BOOT_CLOSURE_HAS_REST_ARGS (closure))
     args = scm_cons_star (scm_from_locale_symbol ("_"), args);
+  /* FIXME: optionals and rests */
   scm_display (args, port);
   scm_putc ('>', port);
   return 1;
