@@ -25,6 +25,7 @@
 #  include <config.h>
 #endif
 
+#include <stdlib.h>
 #include <stdio.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -54,6 +55,7 @@
 #include "libguile/error.h"
 #include "libguile/eval.h"
 #include "libguile/evalext.h"
+#include "libguile/expand.h"
 #include "libguile/feature.h"
 #include "libguile/filesys.h"
 #include "libguile/fluids.h"
@@ -133,7 +135,6 @@
 #include "libguile/extensions.h"
 #include "libguile/uniform.h"
 #include "libguile/srfi-4.h"
-#include "libguile/discouraged.h"
 #include "libguile/deprecated.h"
 
 #include "libguile/init.h"
@@ -160,7 +161,7 @@ fixconfig (char *s1, char *s2, int s)
   fputs ("\nin ", stderr);
   fputs (s ? "setjump" : "scmfig", stderr);
   fputs (".h and recompile scm\n", stderr);
-  exit (1);
+  exit (EXIT_FAILURE);
 }
 
 
@@ -326,8 +327,8 @@ static void *invoke_main_func(void *body_data);
    Call MAIN_FUNC, passing it CLOSURE, ARGC, and ARGV.  MAIN_FUNC
    should do all the work of the program (initializing other packages,
    reading user input, etc.) before returning.  When MAIN_FUNC
-   returns, call exit (0); this function never returns.  If you want
-   some other exit value, MAIN_FUNC may call exit itself.
+   returns, call exit (EXIT_FAILURE); this function never returns.
+   If you want some other exit value, MAIN_FUNC may call exit itself.
 
    scm_boot_guile arranges for program-arguments to return the strings
    given by ARGC and ARGV.  If MAIN_FUNC modifies ARGC/ARGV, should
@@ -368,7 +369,7 @@ scm_boot_guile (int argc, char ** argv, void (*main_func) (), void *closure)
   if (res == NULL)
     exit (EXIT_FAILURE);
   else
-    exit (0);
+    exit (EXIT_SUCCESS);
 }
 
 static void *
@@ -546,6 +547,7 @@ scm_i_init_guile (SCM_STACKITEM *base)
   scm_init_guardians (); /* requires smob_prehistory */
   scm_init_vports ();
   scm_init_standard_ports ();  /* Requires fports */
+  scm_init_expand ();   /* Requires structs */
   scm_init_memoize ();  /* Requires smob_prehistory */
   scm_init_eval ();     /* Requires smob_prehistory */
   scm_init_load_path ();
@@ -560,16 +562,10 @@ scm_i_init_guile (SCM_STACKITEM *base)
 
   scm_init_goops ();
 
-#if SCM_ENABLE_DISCOURAGED == 1
-  scm_i_init_discouraged ();
-#endif
-
 #if SCM_ENABLE_DEPRECATED == 1
   scm_i_init_deprecated ();
 #endif
 
-  scm_init_threads_default_dynamic_state ();
-
   scm_initialized_p = 1;
 
 #ifdef STACK_CHECKING
@@ -583,6 +579,10 @@ scm_i_init_guile (SCM_STACKITEM *base)
   atexit (cleanup_for_exit);
   scm_load_startup_files ();
   scm_init_load_should_autocompile ();
+
+  /* Capture the dynamic state after loading boot-9, so that new threads end up
+     in the guile-user module. */
+  scm_init_threads_default_dynamic_state ();
 }
 
 /*
