@@ -56,8 +56,6 @@ SCM_API SCM scm_program_bindings (SCM program);
 SCM_API SCM scm_program_sources (SCM program);
 SCM_API SCM scm_program_source (SCM program, SCM ip);
 SCM_API SCM scm_program_arities (SCM program);
-SCM_API SCM scm_program_properties (SCM program);
-SCM_API SCM scm_program_name (SCM program);
 SCM_API SCM scm_program_objects (SCM program);
 SCM_API SCM scm_program_module (SCM program);
 SCM_API SCM scm_program_num_free_variables (SCM program);
@@ -67,6 +65,7 @@ SCM_API SCM scm_program_objcode (SCM program);
 
 SCM_API SCM scm_c_program_source (SCM program, size_t ip);
 
+SCM_INTERNAL SCM scm_i_program_properties (SCM program);
 SCM_INTERNAL int scm_i_program_arity (SCM program, int *req, int *opt, int *rest);
 SCM_INTERNAL void scm_i_program_print (SCM program, SCM port,
                                        scm_print_state *pstate);
