@@ -189,6 +189,7 @@ static SCM *scm_read_hash_procedures;
 static inline SCM scm_read_scsh_block_comment (scm_t_wchar, SCM);
 static SCM scm_read_r6rs_block_comment (scm_t_wchar, SCM);
 static SCM scm_read_commented_expression (scm_t_wchar, SCM);
+static SCM scm_read_shebang (scm_t_wchar, SCM);
 static SCM scm_get_hash_procedure (int);
 
 /* Read from PORT until a delimiter (e.g., a whitespace) is read.  Put the
@@ -309,7 +310,7 @@ flush_ws (SCM port, const char *eoferr)
 	    eoferr = "read_sharp";
 	    goto goteof;
 	  case '!':
-	    scm_read_scsh_block_comment (c, port);
+	    scm_read_shebang (c, port);
 	    break;
 	  case ';':
 	    scm_read_commented_expression (c, port);
@@ -391,9 +392,15 @@ scm_read_sexp (scm_t_wchar chr, SCM port)
     {
       SCM new_tail;
 
+      if (c == ')' || (SCM_SQUARE_BRACKETS_P && c == ']'))
+        scm_i_input_error (FUNC_NAME, port,
+                           "in pair: mismatched close paren: ~A",
+                           scm_list_1 (SCM_MAKE_CHAR (c)));
+
       scm_ungetc (c, port);
-      if (scm_is_eq (scm_sym_dot,
-		     (tmp = scm_read_expression (port))))
+      tmp = scm_read_expression (port);
+
+      if (scm_is_eq (scm_sym_dot, tmp))
 	{
 	  SCM_SETCDR (tl, tmp = scm_read_expression (port));
 
@@ -959,16 +966,16 @@ scm_read_character (scm_t_wchar chr, SCM port)
           if (SCM_IS_UNICODE_CHAR (c))
             return SCM_MAKE_CHAR (c);
           else
-            scm_i_input_error (FUNC_NAME, port, 
+            scm_i_input_error (FUNC_NAME, port,
                                "out-of-range octal character escape: ~a",
                                scm_list_1 (charname));
         }
     }
 
-  if (cp == 'x' && (charname_len > 1) && SCM_R6RS_ESCAPES_P)
+  if (cp == 'x' && (charname_len > 1))
     {
       SCM p;
-      
+
       /* Convert from hex, skipping the initial 'x' character in CHARNAME */
       p = scm_string_to_number (scm_c_substring (charname, 1, charname_len),
                                 scm_from_uint (16));
@@ -1109,6 +1116,40 @@ scm_read_scsh_block_comment (scm_t_wchar chr, SCM port)
   return SCM_UNSPECIFIED;
 }
 
+static inline SCM
+scm_read_shebang (scm_t_wchar chr, SCM port)
+{
+  int c = 0;
+  if ((c = scm_get_byte_or_eof (port)) != 'r')
+    {
+      scm_ungetc (c, port);
+      return scm_read_scsh_block_comment (chr, port);
+    }
+  if ((c = scm_get_byte_or_eof (port)) != '6')
+    {
+      scm_ungetc (c, port);
+      scm_ungetc ('r', port);
+      return scm_read_scsh_block_comment (chr, port);
+    }
+  if ((c = scm_get_byte_or_eof (port)) != 'r')
+    {
+      scm_ungetc (c, port);
+      scm_ungetc ('6', port);
+      scm_ungetc ('r', port);
+      return scm_read_scsh_block_comment (chr, port);
+    }
+  if ((c = scm_get_byte_or_eof (port)) != 's')
+    {
+      scm_ungetc (c, port);
+      scm_ungetc ('r', port);
+      scm_ungetc ('6', port);
+      scm_ungetc ('r', port);
+      return scm_read_scsh_block_comment (chr, port);
+    }
+  
+  return SCM_UNSPECIFIED;
+}
+
 static SCM
 scm_read_r6rs_block_comment (scm_t_wchar chr, SCM port)
 {
@@ -1323,7 +1364,7 @@ scm_read_sharp (scm_t_wchar chr, SCM port)
     case '{':
       return (scm_read_extended_symbol (chr, port));
     case '!':
-      return (scm_read_scsh_block_comment (chr, port));
+      return (scm_read_shebang (chr, port));
     case ';':
       return (scm_read_commented_expression (chr, port));
     case '`':
@@ -1586,11 +1627,10 @@ char *
 scm_i_scan_for_encoding (SCM port)
 {
   char header[SCM_ENCODING_SEARCH_SIZE+1];
-  size_t bytes_read;
+  size_t bytes_read, encoding_length, i;
   char *encoding = NULL;
   int utf8_bom = 0;
-  char *pos;
-  int i;
+  char *pos, *encoding_start;
   int in_comment;
 
   if (SCM_FPORTP (port) && !SCM_FDES_RANDOM_P (SCM_FPORT_FDES (port)))
@@ -1628,46 +1668,50 @@ scm_i_scan_for_encoding (SCM port)
     pos ++;
 
   /* grab the next token */
+  encoding_start = pos;
   i = 0;
-  while (pos + i - header <= SCM_ENCODING_SEARCH_SIZE 
-         && pos + i - header < bytes_read
-	 && (isalnum ((int) pos[i]) || strchr ("_-.:/,+=()", pos[i]) != NULL))
+  while (encoding_start + i - header <= SCM_ENCODING_SEARCH_SIZE
+         && encoding_start + i - header < bytes_read
+	 && (isalnum ((int) encoding_start[i])
+	     || strchr ("_-.:/,+=()", encoding_start[i]) != NULL))
     i++;
 
-  if (i == 0)
+  encoding_length = i;
+  if (encoding_length == 0)
     return NULL;
 
-  encoding = scm_gc_strndup (pos, i, "encoding");
-  for (i = 0; i < strlen (encoding); i++)
+  encoding = scm_gc_strndup (encoding_start, encoding_length, "encoding");
+  for (i = 0; i < encoding_length; i++)
     encoding[i] = toupper ((int) encoding[i]);
 
   /* push backwards to make sure we were in a comment */
   in_comment = 0;
-  while (pos - i - header > 0)
+  pos = encoding_start;
+  while (pos >= header)
     {
-      if (*(pos - i) == '\n')
+      if (*pos == '\n')
 	{
 	  /* This wasn't in a semicolon comment. Check for a
 	   hash-bang comment. */
 	  char *beg = strstr (header, "#!");
 	  char *end = strstr (header, "!#");
-	  if (beg < pos && pos < end)
+	  if (beg < encoding_start && encoding_start + encoding_length < end)
 	    in_comment = 1;
 	  break;
 	}
-      if (*(pos - i) == ';')
+      if (*pos == ';')
 	{
 	  in_comment = 1;
 	  break;
 	}
-      i ++;
+      pos --;
     }
   if (!in_comment)
     /* This wasn't in a comment */
     return NULL;
 
   if (utf8_bom && strcmp(encoding, "UTF-8"))
-    scm_misc_error (NULL, 
+    scm_misc_error (NULL,
 		    "the port input declares the encoding ~s but is encoded as UTF-8",
 		    scm_list_1 (scm_from_locale_string (encoding)));
 
