@@ -22,6 +22,7 @@
 #  include <config.h>
 #endif
 
+#include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
 #include <ctype.h>
@@ -385,7 +386,6 @@ scm_shell_usage (int fatal, char *message)
            "                 Default is to enable autocompilation of source\n"
            "                 files.\n"
 	   "  -q             inhibit loading of user init file\n"
-           "  --emacs        enable Emacs protocol (experimental)\n"
 	   "  --use-srfi=LS  load SRFI modules for the SRFIs in LS,\n"
 	   "                 which is a list of numbers like \"2,13,14\"\n"
            "  -h, --help     display this help and exit\n"
@@ -407,6 +407,7 @@ SCM_SYMBOL (sym_command_line, "command-line");
 SCM_SYMBOL (sym_begin, "begin");
 SCM_SYMBOL (sym_turn_on_debugging, "turn-on-debugging");
 SCM_SYMBOL (sym_load_user_init, "load-user-init");
+SCM_SYMBOL (sym_ice_9, "ice-9");
 SCM_SYMBOL (sym_top_repl, "top-repl");
 SCM_SYMBOL (sym_quit, "quit");
 SCM_SYMBOL (sym_use_srfis, "use-srfis");
@@ -454,7 +455,6 @@ scm_compile_shell_switches (int argc, char **argv)
   SCM user_load_path = SCM_EOL; /* for -L switch */
   int interactive = 1;		/* Should we go interactive when done? */
   int inhibit_user_init = 0;	/* Don't load user init file */
-  int use_emacs_interface = 0;
   int turn_on_debugging = 0;
   int dont_turn_on_debugging = 0;
 
@@ -603,9 +603,6 @@ scm_compile_shell_switches (int argc, char **argv)
         scm_variable_set_x (scm_c_lookup ("%load-should-autocompile"),
                             SCM_BOOL_F);
 
-      else if (! strcmp (argv[i], "--emacs")) /* use emacs protocol */ 
-	use_emacs_interface = 1;
-
       else if (! strcmp (argv[i], "-q")) /* don't load user init */ 
 	inhibit_user_init = 1;
 
@@ -647,7 +644,7 @@ scm_compile_shell_switches (int argc, char **argv)
 	       || ! strcmp (argv[i], "--help"))
 	{
 	  scm_shell_usage (0, 0);
-	  exit (0);
+	  exit (EXIT_SUCCESS);
 	}
 
       else if (! strcmp (argv[i], "-v")
@@ -657,7 +654,7 @@ scm_compile_shell_switches (int argc, char **argv)
 	  version_etc (stdout, scm_usage_name, PACKAGE_NAME, PACKAGE_VERSION,
 		       /* XXX: Use gettext for the string below.  */
 		       "the Guile developers", NULL);
-	  exit (0);
+	  exit (EXIT_SUCCESS);
 	}
 
       else
@@ -676,9 +673,6 @@ scm_compile_shell_switches (int argc, char **argv)
      script/command/whatever.  */
   scm_set_program_arguments (argc ? argc - i : 0, argv + i, argv0);
   
-  /* If the --emacs switch was set, now is when we process it.  */
-  scm_c_define ("use-emacs-interface", scm_from_bool (use_emacs_interface));
-
   /* Handle the `-e' switch, if it was specified.  */
   if (!scm_is_null (entry_point))
     tail = scm_cons (scm_cons2 (entry_point,
@@ -689,7 +683,11 @@ scm_compile_shell_switches (int argc, char **argv)
   /* If we didn't end with a -c or a -s, start the repl.  */
   if (interactive)
     {
-      tail = scm_cons (scm_cons (sym_top_repl, SCM_EOL), tail);
+      tail = scm_cons (scm_list_1 (scm_list_3
+                                   (sym_at,
+                                    scm_list_2 (sym_ice_9, sym_top_repl),
+                                    sym_top_repl)),
+                       tail);
     }
   else
     {
