@@ -53,6 +53,7 @@ VM_NAME (SCM vm, SCM program, SCM *argv, int nargs)
 
   /* Internal variables */
   int nvalues = 0;
+  const char *func_name = NULL;         /* used for error reporting */
   SCM finish_args;                      /* used both for returns: both in error
                                            and normal situations */
 #ifdef HAVE_LABELS_AS_VALUES
@@ -139,12 +140,14 @@ VM_NAME (SCM vm, SCM program, SCM *argv, int nargs)
     goto vm_error;
 
   vm_error_unbound:
+    /* At this point FINISH_ARGS should be a one-element list containing
+       the name of the unbound variable.  */
     err_msg  = scm_from_locale_string ("VM: Unbound variable: ~s");
     goto vm_error;
 
-  vm_error_wrong_type_arg:
-    err_msg  = scm_from_locale_string ("VM: Wrong type argument");
-    finish_args = SCM_EOL;
+  vm_error_apply_to_non_list:
+    scm_error (scm_arg_type_key, "apply", "Apply to non-list: ~S",
+               finish_args, finish_args);
     goto vm_error;
 
   vm_error_kwargs_length_not_even:
@@ -181,13 +184,17 @@ VM_NAME (SCM vm, SCM program, SCM *argv, int nargs)
 
   vm_error_wrong_type_apply:
     SYNC_ALL ();
-    scm_error (scm_arg_type_key, FUNC_NAME, "Wrong type to apply: ~S",
+    scm_error (scm_arg_type_key, NULL, "Wrong type to apply: ~S",
                scm_list_1 (program), scm_list_1 (program));
     goto vm_error;
 
   vm_error_stack_overflow:
     err_msg  = scm_from_locale_string ("VM: Stack overflow");
     finish_args = SCM_EOL;
+    if (stack_limit < vp->stack_base + vp->stack_size)
+      /* There are VM_STACK_RESERVE_SIZE bytes left.  Make them available so
+	 that `throw' below can run on this VM.  */
+      vp->stack_limit = vp->stack_base + vp->stack_size;
     goto vm_error;
 
   vm_error_stack_underflow:
@@ -201,25 +208,25 @@ VM_NAME (SCM vm, SCM program, SCM *argv, int nargs)
 
   vm_error_not_a_pair:
     SYNC_ALL ();
-    scm_wrong_type_arg_msg (FUNC_NAME, 1, finish_args, "pair");
+    scm_wrong_type_arg_msg (func_name, 1, finish_args, "pair");
     /* shouldn't get here */
     goto vm_error;
 
   vm_error_not_a_bytevector:
     SYNC_ALL ();
-    scm_wrong_type_arg_msg (FUNC_NAME, 1, finish_args, "bytevector");
+    scm_wrong_type_arg_msg (func_name, 1, finish_args, "bytevector");
     /* shouldn't get here */
     goto vm_error;
 
   vm_error_not_a_struct:
     SYNC_ALL ();
-    scm_wrong_type_arg_msg (FUNC_NAME, 1, finish_args, "struct");
+    scm_wrong_type_arg_msg (func_name, 1, finish_args, "struct");
     /* shouldn't get here */
     goto vm_error;
 
   vm_error_not_a_thunk:
     SYNC_ALL ();
-    scm_wrong_type_arg_msg (FUNC_NAME, 1, finish_args, "thunk");
+    scm_wrong_type_arg_msg ("dynamic-wind", 1, finish_args, "thunk");
     /* shouldn't get here */
     goto vm_error;
 
