@@ -111,23 +111,24 @@ VM_DEFINE_FUNCTION (138, cons, "cons", 2)
   RETURN (x);
 }
 
-#define VM_VALIDATE_CONS(x)                     \
+#define VM_VALIDATE_CONS(x, proc)		\
   if (SCM_UNLIKELY (!scm_is_pair (x)))          \
-    { finish_args = x;                          \
+    { func_name = proc;                         \
+      finish_args = x;                          \
       goto vm_error_not_a_pair;                 \
     }
   
 VM_DEFINE_FUNCTION (139, car, "car", 1)
 {
   ARGS1 (x);
-  VM_VALIDATE_CONS (x);
+  VM_VALIDATE_CONS (x, "car");
   RETURN (SCM_CAR (x));
 }
 
 VM_DEFINE_FUNCTION (140, cdr, "cdr", 1)
 {
   ARGS1 (x);
-  VM_VALIDATE_CONS (x);
+  VM_VALIDATE_CONS (x, "cdr");
   RETURN (SCM_CDR (x));
 }
 
@@ -136,7 +137,7 @@ VM_DEFINE_INSTRUCTION (141, set_car, "set-car!", 0, 2, 0)
   SCM x, y;
   POP (y);
   POP (x);
-  VM_VALIDATE_CONS (x);
+  VM_VALIDATE_CONS (x, "set-car!");
   SCM_SETCAR (x, y);
   NEXT;
 }
@@ -146,7 +147,7 @@ VM_DEFINE_INSTRUCTION (142, set_cdr, "set-cdr!", 0, 2, 0)
   SCM x, y;
   POP (y);
   POP (x);
-  VM_VALIDATE_CONS (x);
+  VM_VALIDATE_CONS (x, "set-cdr!");
   SCM_SETCDR (x, y);
   NEXT;
 }
@@ -386,6 +387,7 @@ VM_DEFINE_INSTRUCTION (163, make_array, "make-array", 3, -1, 1)
   len = (len << 8) + FETCH ();
   POP (shape);
   SYNC_REGISTER ();
+  PRE_CHECK_UNDERFLOW (len);
   ret = scm_from_contiguous_array (shape, sp - len + 1, len);
   DROPN (len);
   PUSH (ret);
@@ -396,9 +398,10 @@ VM_DEFINE_INSTRUCTION (163, make_array, "make-array", 3, -1, 1)
 /*
  * Structs
  */
-#define VM_VALIDATE_STRUCT(obj)			\
+#define VM_VALIDATE_STRUCT(obj, proc)           \
   if (SCM_UNLIKELY (!SCM_STRUCTP (obj)))	\
     {						\
+      func_name = proc;                         \
       finish_args = (obj);			\
       goto vm_error_not_a_struct;		\
     }
@@ -412,7 +415,7 @@ VM_DEFINE_FUNCTION (164, struct_p, "struct?", 1)
 VM_DEFINE_FUNCTION (165, struct_vtable, "struct-vtable", 1)
 {
   ARGS1 (obj);
-  VM_VALIDATE_STRUCT (obj);
+  VM_VALIDATE_STRUCT (obj, "struct_vtable");
   RETURN (SCM_STRUCT_VTABLE (obj));
 }
 
@@ -420,36 +423,33 @@ VM_DEFINE_INSTRUCTION (166, make_struct, "make-struct", 2, -1, 1)
 {
   unsigned h = FETCH ();
   unsigned l = FETCH ();
-  scm_t_bits n_args = ((h << 8U) + l);
-  SCM vtable = sp[1 - n_args], n_tail = sp[2 - n_args];
-  const SCM *inits = sp - n_args + 3;
-
-  sp -= n_args - 1;
+  scm_t_bits n = ((h << 8U) + l);
+  SCM vtable = sp[-(n - 1)];
+  const SCM *inits = sp - n + 2;
+  SCM ret;
 
   SYNC_REGISTER ();
 
   if (SCM_LIKELY (SCM_STRUCTP (vtable)
   		  && SCM_VTABLE_FLAG_IS_SET (vtable, SCM_VTABLE_FLAG_SIMPLE)
-		  && SCM_I_INUMP (n_tail)))
+                  && (SCM_STRUCT_DATA_REF (vtable, scm_vtable_index_size) + 1
+                      == n)
+                  && !SCM_VTABLE_INSTANCE_FINALIZER (vtable)))
     {
-      scm_t_bits n_inits, len;
-
-      n_inits = SCM_I_INUM (n_tail) + n_args - 2;
-      len = SCM_STRUCT_DATA_REF (vtable, scm_vtable_index_size);
-
-      if (SCM_LIKELY (n_inits == len))
-	{
-	  SCM obj;
-
-	  obj = scm_i_alloc_struct (SCM_STRUCT_DATA (vtable), n_inits);
-	  memcpy (SCM_STRUCT_DATA (obj), inits, n_inits * sizeof (SCM));
-
-	  RETURN (obj);
-	}
+      /* Verily, we are making a simple struct with the right number of
+         initializers, and no finalizer. */
+      ret = scm_words ((scm_t_bits)SCM_STRUCT_DATA (vtable) | scm_tc3_struct,
+                       n + 1);
+      SCM_SET_CELL_WORD_1 (ret, (scm_t_bits)SCM_CELL_OBJECT_LOC (ret, 2));
+      memcpy (SCM_STRUCT_DATA (ret), inits, (n - 1) * sizeof (SCM));
     }
+  else
+    ret = scm_c_make_structv (vtable, 0, n - 1, (scm_t_bits *) inits);
 
-  RETURN (scm_c_make_structv (vtable, scm_to_size_t (n_tail),
-			      n_args - 2, (scm_t_bits *) inits));
+  DROPN (n);
+  PUSH (ret);
+
+  NEXT;
 }
 
 VM_DEFINE_FUNCTION (167, struct_ref, "struct-ref", 2)
@@ -545,11 +545,17 @@ VM_DEFINE_INSTRUCTION (171, slot_set, "slot-set", 0, 3, 0)
 /*
  * Bytevectors
  */
-#define VM_VALIDATE_BYTEVECTOR(x)               \
-  if (SCM_UNLIKELY (!SCM_BYTEVECTOR_P (x)))     \
-    { finish_args = x;                          \
-      goto vm_error_not_a_bytevector;           \
-    }
+#define VM_VALIDATE_BYTEVECTOR(x, proc)		\
+  do						\
+    {						\
+      if (SCM_UNLIKELY (!SCM_BYTEVECTOR_P (x)))	\
+	{					\
+          func_name = proc;                     \
+	  finish_args = x;			\
+	  goto vm_error_not_a_bytevector;	\
+	}					\
+    }						\
+  while (0)
 
 #define BV_REF_WITH_ENDIANNESS(stem, fn_stem)                           \
 {                                                                       \
@@ -564,6 +570,10 @@ VM_DEFINE_INSTRUCTION (171, slot_set, "slot-set", 0, 3, 0)
   }                                                                     \
 }
 
+/* Return true (non-zero) if PTR has suitable alignment for TYPE.  */
+#define ALIGNED_P(ptr, type)			\
+  ((scm_t_uintptr) (ptr) % alignof (type) == 0)
+
 VM_DEFINE_FUNCTION (172, bv_u16_ref, "bv-u16-ref", 3)
 BV_REF_WITH_ENDIANNESS (u16, u16)
 VM_DEFINE_FUNCTION (173, bv_s16_ref, "bv-s16-ref", 3)
@@ -585,15 +595,19 @@ BV_REF_WITH_ENDIANNESS (f64, ieee_double)
 
 #define BV_FIXABLE_INT_REF(stem, fn_stem, type, size)			\
 {									\
-  long i = 0;								\
+  long i;								\
+  const scm_t_ ## type *int_ptr;					\
   ARGS2 (bv, idx);							\
-  VM_VALIDATE_BYTEVECTOR (bv);						\
+									\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-ref");                      \
+  i = SCM_I_INUM (idx);							\
+  int_ptr = (scm_t_ ## type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);	\
+									\
   if (SCM_LIKELY (SCM_I_INUMP (idx)					\
-                  && ((i = SCM_I_INUM (idx)) >= 0)			\
+                  && (i >= 0)						\
                   && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))		\
-                  && (i % size == 0)))					\
-    RETURN (SCM_I_MAKINUM (*(scm_t_##type*)				\
-                           (SCM_BYTEVECTOR_CONTENTS (bv) + i)));	\
+                  && (ALIGNED_P (int_ptr, scm_t_ ## type))))		\
+    RETURN (SCM_I_MAKINUM (*int_ptr));					\
   else									\
     {									\
       SYNC_REGISTER ();							\
@@ -603,14 +617,20 @@ BV_REF_WITH_ENDIANNESS (f64, ieee_double)
 
 #define BV_INT_REF(stem, type, size)					\
 {									\
-  long i = 0;								\
+  long i;								\
+  const scm_t_ ## type *int_ptr;					\
   ARGS2 (bv, idx);							\
-  VM_VALIDATE_BYTEVECTOR (bv);						\
+									\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-ref");                      \
+  i = SCM_I_INUM (idx);							\
+  int_ptr = (scm_t_ ## type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);	\
+									\
   if (SCM_LIKELY (SCM_I_INUMP (idx)					\
-                  && ((i = SCM_I_INUM (idx)) >= 0)			\
+                  && (i >= 0)						\
                   && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))		\
-                  && (i % size == 0)))					\
-    { scm_t_##type x = (*(scm_t_##type*)(SCM_BYTEVECTOR_CONTENTS (bv) + i)); \
+                  && (ALIGNED_P (int_ptr, scm_t_ ## type))))		\
+    {									\
+      scm_t_ ## type x = *int_ptr;					\
       if (SCM_FIXABLE (x))						\
         RETURN (SCM_I_MAKINUM (x));					\
       else								\
@@ -628,15 +648,20 @@ BV_REF_WITH_ENDIANNESS (f64, ieee_double)
 
 #define BV_FLOAT_REF(stem, fn_stem, type, size)				\
 {									\
-  long i = 0;								\
+  long i;								\
+  const type *float_ptr;						\
   ARGS2 (bv, idx);							\
-  VM_VALIDATE_BYTEVECTOR (bv);						\
+									\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-ref");                      \
+  i = SCM_I_INUM (idx);							\
+  float_ptr = (type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);		\
+									\
   SYNC_REGISTER ();							\
   if (SCM_LIKELY (SCM_I_INUMP (idx)					\
-                  && ((i = SCM_I_INUM (idx)) >= 0)			\
+                  && (i >= 0)						\
                   && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))		\
-                  && (i % size == 0)))					\
-    RETURN (scm_from_double ((*(type*)(SCM_BYTEVECTOR_CONTENTS (bv) + i)))); \
+                  && (ALIGNED_P (float_ptr, type))))			\
+    RETURN (scm_from_double (*float_ptr));				\
   else									\
     RETURN (scm_bytevector_ ## fn_stem ## _native_ref (bv, idx));	\
 }
@@ -708,52 +733,70 @@ BV_SET_WITH_ENDIANNESS (f64, ieee_double)
 
 #undef BV_SET_WITH_ENDIANNESS
 
-#define BV_FIXABLE_INT_SET(stem, fn_stem, type, min, max, size)         \
-{                                                                       \
-  long i = 0, j = 0;                                                    \
-  SCM bv, idx, val; POP (val); POP (idx); POP (bv);                     \
-  VM_VALIDATE_BYTEVECTOR (bv);                                          \
-  if (SCM_LIKELY (SCM_I_INUMP (idx)                                     \
-                  && ((i = SCM_I_INUM (idx)) >= 0)                      \
-                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))           \
-                  && (i % size == 0)                                    \
-                  && (SCM_I_INUMP (val))                                \
-                  && ((j = SCM_I_INUM (val)) >= min)                    \
-                  && (j <= max)))                                       \
-    *(scm_t_##type*) (SCM_BYTEVECTOR_CONTENTS (bv) + i) = (scm_t_##type)j; \
-  else                                                                  \
-    scm_bytevector_##fn_stem##_set_x (bv, idx, val);                    \
-  NEXT;                                                                 \
+#define BV_FIXABLE_INT_SET(stem, fn_stem, type, min, max, size)		\
+{									\
+  long i, j = 0;							\
+  SCM bv, idx, val;							\
+  scm_t_ ## type *int_ptr;						\
+									\
+  POP (val); POP (idx); POP (bv);					\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-set");                      \
+  i = SCM_I_INUM (idx);							\
+  int_ptr = (scm_t_ ## type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);	\
+									\
+  if (SCM_LIKELY (SCM_I_INUMP (idx)					\
+                  && (i >= 0)						\
+                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))		\
+                  && (ALIGNED_P (int_ptr, scm_t_ ## type))		\
+                  && (SCM_I_INUMP (val))				\
+                  && ((j = SCM_I_INUM (val)) >= min)			\
+                  && (j <= max)))					\
+    *int_ptr = (scm_t_ ## type) j;					\
+  else									\
+    scm_bytevector_ ## fn_stem ## _set_x (bv, idx, val);		\
+  NEXT;									\
 }
 
-#define BV_INT_SET(stem, type, size)                                    \
-{                                                                       \
-  long i = 0;                                                           \
-  SCM bv, idx, val; POP (val); POP (idx); POP (bv);                     \
-  VM_VALIDATE_BYTEVECTOR (bv);                                          \
-  if (SCM_LIKELY (SCM_I_INUMP (idx)                                     \
-                  && ((i = SCM_I_INUM (idx)) >= 0)                      \
-                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))           \
-                  && (i % size == 0)))                                  \
-    *(scm_t_##type*) (SCM_BYTEVECTOR_CONTENTS (bv) + i) = scm_to_##type (val); \
-  else                                                                  \
-    scm_bytevector_##stem##_native_set_x (bv, idx, val);                \
-  NEXT;                                                                 \
+#define BV_INT_SET(stem, type, size)					\
+{									\
+  long i = 0;								\
+  SCM bv, idx, val;							\
+  scm_t_ ## type *int_ptr;						\
+									\
+  POP (val); POP (idx); POP (bv);					\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-set");                      \
+  i = SCM_I_INUM (idx);							\
+  int_ptr = (scm_t_ ## type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);	\
+									\
+  if (SCM_LIKELY (SCM_I_INUMP (idx)					\
+                  && (i >= 0)						\
+                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))		\
+                  && (ALIGNED_P (int_ptr, scm_t_ ## type))))		\
+    *int_ptr = scm_to_ ## type (val);					\
+  else									\
+    scm_bytevector_ ## stem ## _native_set_x (bv, idx, val);		\
+  NEXT;									\
 }
 
-#define BV_FLOAT_SET(stem, fn_stem, type, size)                         \
-{                                                                       \
-  long i = 0;                                                           \
-  SCM bv, idx, val; POP (val); POP (idx); POP (bv);                     \
-  VM_VALIDATE_BYTEVECTOR (bv);                                          \
-  if (SCM_LIKELY (SCM_I_INUMP (idx)                                     \
-                  && ((i = SCM_I_INUM (idx)) >= 0)                      \
-                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))           \
-                  && (i % size == 0)))                                  \
-    *(type*) (SCM_BYTEVECTOR_CONTENTS (bv) + i) = scm_to_double (val);  \
-  else                                                                  \
-    scm_bytevector_##fn_stem##_native_set_x (bv, idx, val);             \
-  NEXT;                                                                 \
+#define BV_FLOAT_SET(stem, fn_stem, type, size)			\
+{								\
+  long i = 0;							\
+  SCM bv, idx, val;						\
+  type *float_ptr;						\
+								\
+  POP (val); POP (idx); POP (bv);				\
+  VM_VALIDATE_BYTEVECTOR (bv, "bv-" #stem "-set");              \
+  i = SCM_I_INUM (idx);						\
+  float_ptr = (type *) (SCM_BYTEVECTOR_CONTENTS (bv) + i);	\
+								\
+  if (SCM_LIKELY (SCM_I_INUMP (idx)				\
+                  && (i >= 0)					\
+                  && (i + size <= SCM_BYTEVECTOR_LENGTH (bv))	\
+                  && (ALIGNED_P (float_ptr, type))))		\
+    *float_ptr = scm_to_double (val);				\
+  else								\
+    scm_bytevector_ ## fn_stem ## _native_set_x (bv, idx, val);	\
+  NEXT;								\
 }
 
 VM_DEFINE_INSTRUCTION (198, bv_u8_set, "bv-u8-set", 0, 3, 0)
