@@ -22,6 +22,7 @@
 
 #include <stdlib.h>
 #include <alloca.h>
+#include <alignof.h>
 #include <string.h>
 
 #include "libguile/bdw-gc.h"
@@ -65,6 +66,10 @@
    for a discussion.  */
 #define VM_ENABLE_PRECISE_STACK_GC_SCAN
 
+/* Size in SCM objects of the stack reserve.  The reserve is used to run
+   exception handling code in case of a VM stack overflow.  */
+#define VM_STACK_RESERVE_SIZE  512
+
 
 
 /*
@@ -185,7 +190,9 @@ vm_dispatch_hook (SCM vm, int hook_num)
 {
   struct scm_vm *vp;
   SCM hook;
-  SCM frame;
+  struct scm_frame c_frame;
+  scm_t_aligned_cell frame;
+  SCM args[1];
 
   vp = SCM_VM_DATA (vm);
   hook = vp->hooks[hook_num];
@@ -193,10 +200,28 @@ vm_dispatch_hook (SCM vm, int hook_num)
   if (SCM_LIKELY (scm_is_false (hook))
       || scm_is_null (SCM_HOOK_PROCEDURES (hook)))
     return;
-  
+
   vp->trace_level--;
-  frame = scm_c_make_frame (vm, vp->fp, vp->sp, vp->ip, 0);
-  scm_c_run_hookn (hook, &frame, 1);
+
+  /* Allocate a frame object on the stack.  This is more efficient than calling
+     `scm_c_make_frame ()' to allocate on the heap, but it forces hooks to not
+     capture frame objects.
+
+     At the same time, procedures such as `frame-procedure' make sense only
+     while the stack frame represented by the frame object is visible, so it
+     seems reasonable to limit the lifetime of frame objects.  */
+
+  c_frame.stack_holder = vm;
+  c_frame.fp = vp->fp;
+  c_frame.sp = vp->sp;
+  c_frame.ip = vp->ip;
+  c_frame.offset = 0;
+  frame.cell.word_0 = SCM_PACK (scm_tc7_frame);
+  frame.cell.word_1 = PTR2SCM (&c_frame);
+  args[0] = PTR2SCM (&frame);
+
+  scm_c_run_hookn (hook, args, 1);
+
   vp->trace_level++;
 }
 
@@ -315,7 +340,24 @@ static SCM sym_vm_run, sym_vm_error, sym_keyword_argument_error, sym_debug;
 void
 scm_i_vm_print (SCM x, SCM port, scm_print_state *pstate)
 {
+  const struct scm_vm *vm;
+
+  vm = SCM_VM_DATA (x);
+
   scm_puts ("#<vm ", port);
+  switch (vm->engine)
+    {
+    case SCM_VM_REGULAR_ENGINE:
+      scm_puts ("regular-engine ", port);
+      break;
+
+    case SCM_VM_DEBUG_ENGINE:
+      scm_puts ("debug-engine ", port);
+      break;
+
+    default:
+      scm_puts ("unknown-engine ", port);
+    }
   scm_uintprint (SCM_UNPACK (x), 16, port);
   scm_puts (">", port);
 }
@@ -464,7 +506,7 @@ make_vm (void)
 #ifdef VM_ENABLE_STACK_NULLING
   memset (vp->stack_base, 0, vp->stack_size * sizeof (SCM));
 #endif
-  vp->stack_limit = vp->stack_base + vp->stack_size;
+  vp->stack_limit = vp->stack_base + vp->stack_size - VM_STACK_RESERVE_SIZE;
   vp->ip    	  = NULL;
   vp->sp    	  = vp->stack_base - 1;
   vp->fp    	  = NULL;
@@ -493,8 +535,7 @@ vm_stack_mark (GC_word *addr, struct GC_ms_entry *mark_stack_ptr,
   vm = * ((struct scm_vm **) addr);
 
   if (vm == NULL
-      || (SCM *) addr != vm->stack_base - 1
-      || vm->stack_limit - vm->stack_base != vm->stack_size)
+      || (SCM *) addr != vm->stack_base - 1)
     /* ADDR must be a pointer to a free-list element, which we must ignore
        (see warning in <gc/gc_mark.h>).  */
     return mark_stack_ptr;
