@@ -1,6 +1,6 @@
 ;;;; Guile Debugger command loop
 
-;;; Copyright (C) 1999, 2001, 2002, 2003, 2006 Free Software Foundation, Inc.
+;;; Copyright (C) 1999, 2001, 2002, 2003, 2006, 2010 Free Software Foundation, Inc.
 ;;;
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -21,6 +21,7 @@
   #:use-module (ice-9 debugger)
   #:use-module (ice-9 debugger state)
   #:use-module (ice-9 debugging traps)
+  #:use-module (ice-9 save-stack)
   #:export (debugger-command-loop
 	    debugger-command-loop-error
 	    debugger-command-loop-quit)
@@ -45,8 +46,6 @@
   (case key
     ((exit-debugger) #f)
     ((signal)
-     ;; Restore stack
-     (fluid-set! the-last-stack (fluid-ref before-signal-stack))
      (apply display-error #f (current-error-port) args))
     (else
      (display "Internal debugger error:\n")
