@@ -26,6 +26,7 @@
 (define-module (ice-9 debugging traps)
   #:use-module (ice-9 regex)
   #:use-module (ice-9 weak-vector)
+  #:use-module (ice-9 scm-style-repl)
   #:use-module (oop goops)
   #:use-module (oop goops describe)
   #:use-module (ice-9 debugging trc)
@@ -494,7 +495,7 @@ it twice."
 	      (or (memq key ignored-keys)
 		  (behaviour (throw->trap-context key
 						  args
-						  pre-unwind-handler-dispatch)))
+						  default-pre-unwind-handler)))
 	      (apply default-default-pre-unwind-handler key args))
 	    default-default-pre-unwind-handler)))
 
