@@ -1,19 +1,19 @@
 ;;;; 	Copyright (C) 2009, 2010 Free Software Foundation, Inc.
-;;;; 
+;;;;
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
 ;;;; License as published by the Free Software Foundation; either
 ;;;; version 3 of the License, or (at your option) any later version.
-;;;; 
+;;;;
 ;;;; This library is distributed in the hope that it will be useful,
 ;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ;;;; Lesser General Public License for more details.
-;;;; 
+;;;;
 ;;;; You should have received a copy of the GNU Lesser General Public
 ;;;; License along with this library; if not, write to the Free Software
 ;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-;;;; 
+;;;;
 
 
 (define-module (language tree-il)
@@ -39,15 +39,15 @@
             <lambda> lambda? make-lambda lambda-src lambda-meta lambda-body
             <lambda-case> lambda-case? make-lambda-case lambda-case-src
                           lambda-case-req lambda-case-opt lambda-case-rest lambda-case-kw
-                          lambda-case-inits lambda-case-vars
+                          lambda-case-inits lambda-case-gensyms
                           lambda-case-body lambda-case-alternate
-            <let> let? make-let let-src let-names let-vars let-vals let-body
-            <letrec> letrec? make-letrec letrec-src letrec-names letrec-vars letrec-vals letrec-body
-            <fix> fix? make-fix fix-src fix-names fix-vars fix-vals fix-body
+            <let> let? make-let let-src let-names let-gensyms let-vals let-body
+            <letrec> letrec? make-letrec letrec-src letrec-in-order? letrec-names letrec-gensyms letrec-vals letrec-body
+            <fix> fix? make-fix fix-src fix-names fix-gensyms fix-vals fix-body
             <let-values> let-values? make-let-values let-values-src let-values-exp let-values-body
             <dynwind> dynwind? make-dynwind dynwind-src dynwind-winder dynwind-body dynwind-unwinder
             <dynlet> dynlet? make-dynlet dynlet-src dynlet-fluids dynlet-vals dynlet-body
-            <dynref> dynref? make-dynref dynref-src dynref-fluid 
+            <dynref> dynref? make-dynref dynref-src dynref-fluid
             <dynset> dynset? make-dynset dynset-src dynset-fluid dynset-exp
             <prompt> prompt? make-prompt prompt-src prompt-tag prompt-body prompt-handler
             <abort> abort? make-abort abort-src abort-tag abort-args abort-tail
@@ -61,33 +61,80 @@
             post-order!
             pre-order!))
 
-(define-type (<tree-il> #:common-slots (src))
-  (<void>)
-  (<const> exp)
-  (<primitive-ref> name)
-  (<lexical-ref> name gensym)
-  (<lexical-set> name gensym exp)
-  (<module-ref> mod name public?)
-  (<module-set> mod name public? exp)
-  (<toplevel-ref> name)
-  (<toplevel-set> name exp)
-  (<toplevel-define> name exp)
-  (<conditional> test consequent alternate)
-  (<application> proc args)
-  (<sequence> exps)
-  (<lambda> meta body)
-  (<lambda-case> req opt rest kw inits vars body alternate)
-  (<let> names vars vals body)
-  (<letrec> names vars vals body)
-  (<fix> names vars vals body)
+(define (print-tree-il exp port)
+  (format port "#<tree-il ~a>" (unparse-tree-il exp)))
+
+(define-syntax borrow-core-vtables
+  (lambda (x)
+    (syntax-case x ()
+      ((_)
+       (let lp ((n 0) (out '()))
+         (if (< n (vector-length %expanded-vtables))
+             (lp (1+ n)
+                 (let* ((vtable (vector-ref %expanded-vtables n))
+                        (stem (struct-ref vtable (+ vtable-offset-user 0)))
+                        (fields (struct-ref vtable (+ vtable-offset-user 2)))
+                        (sfields (map
+                                  (lambda (f) (datum->syntax x f))
+                                  fields))
+                        (type (datum->syntax x (symbol-append '< stem '>)))
+                        (ctor (datum->syntax x (symbol-append 'make- stem)))
+                        (pred (datum->syntax x (symbol-append stem '?))))
+                   (let lp ((n 0) (fields fields)
+                            (out (cons*
+                                  #`(define (#,ctor #,@sfields)
+                                      (make-struct #,type 0 #,@sfields))
+                                  #`(define (#,pred x)
+                                      (and (struct? x)
+                                           (eq? (struct-vtable x) #,type)))
+                                  #`(struct-set! #,type vtable-index-printer
+                                                 print-tree-il)
+                                  #`(define #,type
+                                           (vector-ref %expanded-vtables #,n))
+                                       out)))
+                     (if (null? fields)
+                         out
+                         (lp (1+ n)
+                             (cdr fields)
+                             (let ((acc (datum->syntax
+                                         x (symbol-append stem '- (car fields)))))
+                               (cons #`(define #,acc
+                                         (make-procedure-with-setter
+                                          (lambda (x) (struct-ref x #,n))
+                                          (lambda (x v) (struct-set! x #,n v))))
+                                     out)))))))
+             #`(begin #,@(reverse out))))))))
+
+(borrow-core-vtables)
+
+  ;; (<void>)
+  ;; (<const> exp)
+  ;; (<primitive-ref> name)
+  ;; (<lexical-ref> name gensym)
+  ;; (<lexical-set> name gensym exp)
+  ;; (<module-ref> mod name public?)
+  ;; (<module-set> mod name public? exp)
+  ;; (<toplevel-ref> name)
+  ;; (<toplevel-set> name exp)
+  ;; (<toplevel-define> name exp)
+  ;; (<conditional> test consequent alternate)
+  ;; (<application> proc args)
+  ;; (<sequence> exps)
+  ;; (<lambda> meta body)
+  ;; (<lambda-case> req opt rest kw inits gensyms body alternate)
+  ;; (<let> names gensyms vals body)
+  ;; (<letrec> in-order? names gensyms vals body)
+  ;; (<dynlet> fluids vals body)
+
+(define-type (<tree-il> #:common-slots (src) #:printer print-tree-il)
+  (<fix> names gensyms vals body)
   (<let-values> exp body)
   (<dynwind> winder body unwinder)
-  (<dynlet> fluids vals body)
   (<dynref> fluid)
   (<dynset> fluid exp)
   (<prompt> tag body handler)
   (<abort> tag args tail))
-  
+
 
 
 (define (location x)
@@ -147,15 +194,15 @@
      ((lambda ,meta ,body)
       (make-lambda loc meta (retrans body)))
 
-     ((lambda-case ((,req ,opt ,rest ,kw ,inits ,vars) ,body) ,alternate)
-      (make-lambda-case loc req opt rest kw 
-                        (map retrans inits) vars
+     ((lambda-case ((,req ,opt ,rest ,kw ,inits ,gensyms) ,body) ,alternate)
+      (make-lambda-case loc req opt rest kw
+                        (map retrans inits) gensyms
                         (retrans body)
                         (and=> alternate retrans)))
 
-     ((lambda-case ((,req ,opt ,rest ,kw ,inits ,vars) ,body))
+     ((lambda-case ((,req ,opt ,rest ,kw ,inits ,gensyms) ,body))
       (make-lambda-case loc req opt rest kw
-                        (map retrans inits) vars
+                        (map retrans inits) gensyms
                         (retrans body)
                         #f))
 
@@ -165,33 +212,36 @@
      ((begin . ,exps)
       (make-sequence loc (map retrans exps)))
 
-     ((let ,names ,vars ,vals ,body)
-      (make-let loc names vars (map retrans vals) (retrans body)))
+     ((let ,names ,gensyms ,vals ,body)
+      (make-let loc names gensyms (map retrans vals) (retrans body)))
 
-     ((letrec ,names ,vars ,vals ,body)
-      (make-letrec loc names vars (map retrans vals) (retrans body)))
+     ((letrec ,names ,gensyms ,vals ,body)
+      (make-letrec loc #f names gensyms (map retrans vals) (retrans body)))
 
-     ((fix ,names ,vars ,vals ,body)
-      (make-fix loc names vars (map retrans vals) (retrans body)))
+     ((letrec* ,names ,gensyms ,vals ,body)
+      (make-letrec loc #t names gensyms (map retrans vals) (retrans body)))
+
+     ((fix ,names ,gensyms ,vals ,body)
+      (make-fix loc names gensyms (map retrans vals) (retrans body)))
 
      ((let-values ,exp ,body)
       (make-let-values loc (retrans exp) (retrans body)))
 
      ((dynwind ,winder ,body ,unwinder)
       (make-dynwind loc (retrans winder) (retrans body) (retrans unwinder)))
-     
+
      ((dynlet ,fluids ,vals ,body)
       (make-dynlet loc (map retrans fluids) (map retrans vals) (retrans body)))
-     
+
      ((dynref ,fluid)
       (make-dynref loc (retrans fluid)))
-     
+
      ((dynset ,fluid ,exp)
       (make-dynset loc (retrans fluid) (retrans exp)))
-     
+
      ((prompt ,tag ,body ,handler)
       (make-prompt loc (retrans tag) (retrans body) (retrans handler)))
-     
+
      ((abort ,tag ,args ,tail)
       (make-abort loc (retrans tag) (map retrans args) (retrans tail)))
 
@@ -236,8 +286,8 @@
     ((<lambda> meta body)
      `(lambda ,meta ,(unparse-tree-il body)))
 
-    ((<lambda-case> req opt rest kw inits vars body alternate)
-     `(lambda-case ((,req ,opt ,rest ,kw ,(map unparse-tree-il inits) ,vars)
+    ((<lambda-case> req opt rest kw inits gensyms body alternate)
+     `(lambda-case ((,req ,opt ,rest ,kw ,(map unparse-tree-il inits) ,gensyms)
                     ,(unparse-tree-il body))
                    . ,(if alternate (list (unparse-tree-il alternate)) '())))
 
@@ -247,14 +297,15 @@
     ((<sequence> exps)
      `(begin ,@(map unparse-tree-il exps)))
 
-    ((<let> names vars vals body)
-     `(let ,names ,vars ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
+    ((<let> names gensyms vals body)
+     `(let ,names ,gensyms ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
 
-    ((<letrec> names vars vals body)
-     `(letrec ,names ,vars ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
+    ((<letrec> in-order? names gensyms vals body)
+     `(,(if in-order? 'letrec* 'letrec) ,names ,gensyms
+       ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
 
-    ((<fix> names vars vals body)
-     `(fix ,names ,vars ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
+    ((<fix> names gensyms vals body)
+     `(fix ,names ,gensyms ,(map unparse-tree-il vals) ,(unparse-tree-il body)))
 
     ((<let-values> exp body)
      `(let-values ,(unparse-tree-il exp) ,(unparse-tree-il body)))
@@ -262,20 +313,20 @@
     ((<dynwind> body winder unwinder)
      `(dynwind ,(unparse-tree-il body)
                ,(unparse-tree-il winder) ,(unparse-tree-il unwinder)))
-    
+
     ((<dynlet> fluids vals body)
      `(dynlet ,(map unparse-tree-il fluids) ,(map unparse-tree-il vals)
               ,(unparse-tree-il body)))
-    
+
     ((<dynref> fluid)
      `(dynref ,(unparse-tree-il fluid)))
-    
+
     ((<dynset> fluid exp)
      `(dynref ,(unparse-tree-il fluid) ,(unparse-tree-il exp)))
-    
+
     ((<prompt> tag body handler)
-     `(prompt ,tag ,(unparse-tree-il body) ,(unparse-tree-il handler)))
-    
+     `(prompt ,(unparse-tree-il tag) ,(unparse-tree-il body) ,(unparse-tree-il handler)))
+
     ((<abort> tag args tail)
      `(abort ,(unparse-tree-il tag) ,(map unparse-tree-il args)
              ,(unparse-tree-il tail)))))
@@ -295,58 +346,108 @@
 
     ((<primitive-ref> name)
      name)
-    
+
     ((<lexical-ref> gensym)
      gensym)
-    
+
     ((<lexical-set> gensym exp)
      `(set! ,gensym ,(tree-il->scheme exp)))
-    
+
     ((<module-ref> mod name public?)
      `(,(if public? '@ '@@) ,mod ,name))
-    
+
     ((<module-set> mod name public? exp)
      `(set! (,(if public? '@ '@@) ,mod ,name) ,(tree-il->scheme exp)))
-    
+
     ((<toplevel-ref> name)
      name)
-    
+
     ((<toplevel-set> name exp)
      `(set! ,name ,(tree-il->scheme exp)))
-    
+
     ((<toplevel-define> name exp)
      `(define ,name ,(tree-il->scheme exp)))
-    
+
     ((<lambda> meta body)
      ;; fixme: put in docstring
-     (if (and (lambda-case? body)
-              (not (lambda-case-alternate body)))
-         `(lambda ,@(car (tree-il->scheme body)))
-         `(case-lambda ,@(tree-il->scheme body))))
-    
-    ((<lambda-case> req opt rest kw inits vars body alternate)
-     ;; FIXME! use parse-lambda-case?
-     `((,(if rest (apply cons* vars) vars)
-        ,(tree-il->scheme body))
-       ,@(if alternate (tree-il->scheme alternate) '())))
-    
+     (tree-il->scheme body))
+
+    ((<lambda-case> req opt rest kw inits gensyms body alternate)
+     (cond
+      ((and (not opt) (not kw) (not alternate))
+       `(lambda ,(if rest (apply cons* gensyms) gensyms)
+          ,(tree-il->scheme body)))
+      ((and (not opt) (not kw))
+       (let ((alt-expansion (tree-il->scheme alternate))
+             (formals (if rest (apply cons* gensyms) gensyms)))
+         (case (car alt-expansion)
+           ((lambda)
+            `(case-lambda (,formals ,(tree-il->scheme body))
+                          ,@(cdr alt-expansion)))
+           ((lambda*)
+            `(case-lambda* (,formals ,(tree-il->scheme body))
+                           ,(cdr alt-expansion)))
+           ((case-lambda)
+            `(case-lambda (,formals ,(tree-il->scheme body))
+                          ,@(cdr alt-expansion)))
+           ((case-lambda*)
+            `(case-lambda* (,formals ,(tree-il->scheme body))
+                           ,@(cdr alt-expansion))))))
+      (else
+       (let* ((alt-expansion (and alternate (tree-il->scheme alternate)))
+              (nreq (length req))
+              (nopt (if opt (length opt) 0))
+              (restargs (if rest (list-ref gensyms (+ nreq nopt)) '()))
+              (reqargs (list-head gensyms nreq))
+              (optargs (if opt
+                           `(#:optional
+                             ,@(map list
+                                    (list-head (list-tail gensyms nreq) nopt)
+                                    (map tree-il->scheme
+                                         (list-head inits nopt))))
+                           '()))
+              (kwargs (if kw
+                          `(#:key
+                            ,@(map list
+                                   (map caddr (cdr kw))
+                                   (map tree-il->scheme
+                                        (list-tail inits nopt))
+                                   (map car (cdr kw)))
+                            ,@(if (car kw)
+                                  '(#:allow-other-keys)
+                                  '()))
+                          '()))
+              (formals `(,@reqargs ,@optargs ,@kwargs . ,restargs)))
+         (if (not alt-expansion)
+             `(lambda* ,formals ,(tree-il->scheme body))
+             (case (car alt-expansion)
+               ((lambda lambda*)
+                `(case-lambda* (,formals ,(tree-il->scheme body))
+                               ,(cdr alt-expansion)))
+               ((case-lambda case-lambda*)
+                `(case-lambda* (,formals ,(tree-il->scheme body))
+                               ,@(cdr alt-expansion)))))))))
+
     ((<const> exp)
      (if (and (self-evaluating? exp) (not (vector? exp)))
          exp
          (list 'quote exp)))
-    
+
     ((<sequence> exps)
      `(begin ,@(map tree-il->scheme exps)))
-    
-    ((<let> vars vals body)
-     `(let ,(map list vars (map tree-il->scheme vals)) ,(tree-il->scheme body)))
-    
-    ((<letrec> vars vals body)
-     `(letrec ,(map list vars (map tree-il->scheme vals)) ,(tree-il->scheme body)))
 
-    ((<fix> vars vals body)
-     ;; not a typo, we really do translate back to letrec
-     `(letrec ,(map list vars (map tree-il->scheme vals)) ,(tree-il->scheme body)))
+    ((<let> gensyms vals body)
+     `(let ,(map list gensyms (map tree-il->scheme vals)) ,(tree-il->scheme body)))
+
+    ((<letrec> in-order? gensyms vals body)
+     `(,(if in-order? 'letrec* 'letrec)
+       ,(map list gensyms (map tree-il->scheme vals)) ,(tree-il->scheme body)))
+
+    ((<fix> gensyms vals body)
+     ;; not a typo, we really do translate back to letrec. use letrec* since it
+     ;; doesn't matter, and the naive letrec* transformation does not require an
+     ;; inner let.
+     `(letrec* ,(map list gensyms (map tree-il->scheme vals)) ,(tree-il->scheme body)))
 
     ((<let-values> exp body)
      `(call-with-values (lambda () ,(tree-il->scheme exp))
@@ -356,24 +457,24 @@
      `(dynamic-wind ,(tree-il->scheme winder)
                     (lambda () ,(tree-il->scheme body))
                     ,(tree-il->scheme unwinder)))
-    
+
     ((<dynlet> fluids vals body)
      `(with-fluids ,(map list
                          (map tree-il->scheme fluids)
                          (map tree-il->scheme vals))
         ,(tree-il->scheme body)))
-    
+
     ((<dynref> fluid)
      `(fluid-ref ,(tree-il->scheme fluid)))
-    
+
     ((<dynset> fluid exp)
      `(fluid-set! ,(tree-il->scheme fluid) ,(tree-il->scheme exp)))
-    
+
     ((<prompt> tag body handler)
-     `((@ (ice-9 control) prompt) 
+     `((@ (ice-9 control) prompt)
        ,(tree-il->scheme tag) (lambda () ,(tree-il->scheme body))
        ,(tree-il->scheme handler)))
-    
+
 
     ((<abort> tag args tail)
      `(apply abort ,(tree-il->scheme tag) ,@(map tree-il->scheme args)
@@ -539,76 +640,76 @@ This is an implementation of `foldts' as described by Andy Wingo in
        (set! (conditional-test x) (lp test))
        (set! (conditional-consequent x) (lp consequent))
        (set! (conditional-alternate x) (lp alternate)))
-      
+
       ((<lexical-set> name gensym exp)
        (set! (lexical-set-exp x) (lp exp)))
-      
+
       ((<module-set> mod name public? exp)
        (set! (module-set-exp x) (lp exp)))
-      
+
       ((<toplevel-set> name exp)
        (set! (toplevel-set-exp x) (lp exp)))
-      
+
       ((<toplevel-define> name exp)
        (set! (toplevel-define-exp x) (lp exp)))
-      
+
       ((<lambda> body)
        (set! (lambda-body x) (lp body)))
-      
+
       ((<lambda-case> inits body alternate)
        (set! inits (map lp inits))
        (set! (lambda-case-body x) (lp body))
        (if alternate
            (set! (lambda-case-alternate x) (lp alternate))))
-      
+
       ((<sequence> exps)
        (set! (sequence-exps x) (map lp exps)))
-      
-      ((<let> vars vals body)
+
+      ((<let> gensyms vals body)
        (set! (let-vals x) (map lp vals))
        (set! (let-body x) (lp body)))
-      
-      ((<letrec> vars vals body)
+
+      ((<letrec> gensyms vals body)
        (set! (letrec-vals x) (map lp vals))
        (set! (letrec-body x) (lp body)))
-      
-      ((<fix> vars vals body)
+
+      ((<fix> gensyms vals body)
        (set! (fix-vals x) (map lp vals))
        (set! (fix-body x) (lp body)))
-      
+
       ((<let-values> exp body)
        (set! (let-values-exp x) (lp exp))
        (set! (let-values-body x) (lp body)))
-      
+
       ((<dynwind> body winder unwinder)
        (set! (dynwind-body x) (lp body))
        (set! (dynwind-winder x) (lp winder))
        (set! (dynwind-unwinder x) (lp unwinder)))
-      
+
       ((<dynlet> fluids vals body)
        (set! (dynlet-fluids x) (map lp fluids))
        (set! (dynlet-vals x) (map lp vals))
        (set! (dynlet-body x) (lp body)))
-      
+
       ((<dynref> fluid)
        (set! (dynref-fluid x) (lp fluid)))
-      
+
       ((<dynset> fluid exp)
        (set! (dynset-fluid x) (lp fluid))
        (set! (dynset-exp x) (lp exp)))
-      
+
       ((<prompt> tag body handler)
        (set! (prompt-tag x) (lp tag))
        (set! (prompt-body x) (lp body))
        (set! (prompt-handler x) (lp handler)))
-      
+
       ((<abort> tag args tail)
        (set! (abort-tag x) (lp tag))
        (set! (abort-args x) (map lp args))
        (set! (abort-tail x) (lp tail)))
-      
+
       (else #f))
-    
+
     (or (f x) x)))
 
 (define (pre-order! f x)
@@ -626,7 +727,7 @@ This is an implementation of `foldts' as described by Andy Wingo in
 
         ((<lexical-set> exp)
          (set! (lexical-set-exp x) (lp exp)))
-               
+
         ((<module-set> exp)
          (set! (module-set-exp x) (lp exp)))
 
@@ -667,28 +768,28 @@ This is an implementation of `foldts' as described by Andy Wingo in
          (set! (dynwind-body x) (lp body))
          (set! (dynwind-winder x) (lp winder))
          (set! (dynwind-unwinder x) (lp unwinder)))
-        
+
         ((<dynlet> fluids vals body)
          (set! (dynlet-fluids x) (map lp fluids))
          (set! (dynlet-vals x) (map lp vals))
          (set! (dynlet-body x) (lp body)))
-      
+
         ((<dynref> fluid)
          (set! (dynref-fluid x) (lp fluid)))
-        
+
         ((<dynset> fluid exp)
          (set! (dynset-fluid x) (lp fluid))
          (set! (dynset-exp x) (lp exp)))
-        
+
         ((<prompt> tag body handler)
          (set! (prompt-tag x) (lp tag))
          (set! (prompt-body x) (lp body))
          (set! (prompt-handler x) (lp handler)))
-        
+
         ((<abort> tag args tail)
          (set! (abort-tag x) (lp tag))
          (set! (abort-args x) (map lp args))
          (set! (abort-tail x) (lp tail)))
-        
+
         (else #f))
       x)))
