@@ -20,13 +20,14 @@
 
 (define-module (language tree-il primitives)
   #:use-module (system base pmatch)
-  #:use-module (rnrs bytevector)
+  #:use-module (rnrs bytevectors)
   #:use-module (system base syntax)
   #:use-module (language tree-il)
   #:use-module (srfi srfi-4)
   #:use-module (srfi srfi-16)
   #:export (resolve-primitives! add-interesting-primitive!
-            expand-primitives! effect-free-primitive?))
+            expand-primitives!
+            effect-free-primitive? effect+exception-free-primitive?))
 
 (define *interesting-primitive-names* 
   '(apply @apply
@@ -59,11 +60,11 @@
     vector-ref vector-set!
     variable-ref variable-set!
     variable-bound?
-    ;; args of variable-set are switched; it needs special help
 
     fluid-ref fluid-set!
 
     @prompt call-with-prompt @abort abort-to-prompt
+    make-prompt-tag
 
     struct? struct-vtable make-struct struct-ref struct-set!
 
@@ -119,7 +120,7 @@
     caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr
     cdaaar cdaadr cdadar cdaddr cddaar cddadr cdddar cddddr
     vector-ref
-    struct? struct-vtable make-struct struct-ref
+    struct? struct-vtable make-struct make-struct/no-tail struct-ref
     bytevector-u8-ref bytevector-s8-ref
     bytevector-u16-ref bytevector-u16-native-ref
     bytevector-s16-ref bytevector-s16-native-ref
@@ -130,14 +131,30 @@
     bytevector-ieee-single-ref bytevector-ieee-single-native-ref
     bytevector-ieee-double-ref bytevector-ieee-double-native-ref))
 
+;; Like *effect-free-primitives* above, but further restricted in that they
+;; cannot raise exceptions.
+(define *effect+exception-free-primitives*
+  '(values
+    eq? eqv? equal?
+    not
+    pair? null? list? acons cons cons*
+    list vector
+    struct?))
 
 (define *effect-free-primitive-table* (make-hash-table))
+(define *effect+exceptions-free-primitive-table* (make-hash-table))
 
-(for-each (lambda (x) (hashq-set! *effect-free-primitive-table* x #t))
+(for-each (lambda (x)
+            (hashq-set! *effect-free-primitive-table* x #t))
           *effect-free-primitives*)
+(for-each (lambda (x) 
+            (hashq-set! *effect+exceptions-free-primitive-table* x #t))
+          *effect+exception-free-primitives*)
 
 (define (effect-free-primitive? prim)
   (hashq-ref *effect-free-primitive-table* prim))
+(define (effect+exception-free-primitive? prim)
+  (hashq-ref *effect+exceptions-free-primitive-table* prim))
 
 (define (resolve-primitives! x mod)
   (post-order!
@@ -212,6 +229,9 @@
       exp)
      ((number? exp)
       `(make-const src ,exp))
+     ((not exp)
+      ;; failed match
+      #f)
      (else (error "bad consequent yall" exp))))
   `(hashq-set! *primitive-expand-table*
                ',sym
@@ -313,9 +333,12 @@
 
 (define-primitive-expander values (x) x)
 
-;; swap args
-(define-primitive-expander variable-set! (var val)
-  (variable-set val var))
+(define-primitive-expander make-struct (vtable tail-size . args)
+  (if (and (const? tail-size)
+           (let ((n (const-exp tail-size)))
+             (and (number? n) (exact? n) (zero? n))))
+      (make-struct/no-tail vtable . args)
+      #f))
 
 (define-primitive-expander u8vector-ref (vec i)
   (bytevector-u8-ref vec i))
