@@ -27,8 +27,9 @@
 ;;; Code:
 
 (define-module (rnrs io ports)
-  :re-export (eof-object? port? input-port? output-port?)
-  :export (eof-object
+  #:version (6)
+  #:re-export (eof-object? port? input-port? output-port?)
+  #:export (eof-object
 
            ;; input & output ports
            port-transcoder binary-port? transcoded-port
@@ -38,6 +39,7 @@
 
            ;; input ports
            open-bytevector-input-port
+           open-string-input-port
            make-custom-binary-input-port
 
            ;; binary input
@@ -47,6 +49,7 @@
 
            ;; output ports
            open-bytevector-output-port
+           open-string-output-port
            make-custom-binary-output-port
 
            ;; binary output
@@ -105,4 +108,17 @@ read from/written to in @var{port}."
       (lambda ()
         (close-port port))))
 
+(define (open-string-input-port str)
+  "Open an input port that will read from @var{str}."
+  (with-fluids ((%default-port-encoding "UTF-8"))
+    (open-input-string str)))
+
+(define (open-string-output-port)
+  "Return two values: an output port that will collect characters written to it
+as a string, and a thunk to retrieve the characters associated with that port."
+  (let ((port (with-fluids ((%default-port-encoding "UTF-8"))
+                (open-output-string))))
+    (values port
+            (lambda () (get-output-string port)))))
+
 ;;; ports.scm ends here
