@@ -27,13 +27,34 @@
 
 #include "srfi-1.h"
 
-/* The intent of this file is to gradually replace those Scheme
- * procedures in srfi-1.scm which extends core primitive procedures,
+/* The intent of this file was to gradually replace those Scheme
+ * procedures in srfi-1.scm that extend core primitive procedures,
  * so that using srfi-1 won't have performance penalties.
  *
- * Please feel free to contribute any new replacements!
+ * However, we now prefer to write these procedures in Scheme, let the compiler
+ * optimize them, and have the VM execute them efficiently.
  */
 
+
+/* The `(srfi srfi-1)' module.  */
+static SCM srfi1_module = SCM_BOOL_F;
+
+/* Cache variable NAME in C variable VAR.  */
+#define CACHE_VAR(var, name)						\
+  static SCM var = SCM_BOOL_F;						\
+  if (scm_is_false (var))						\
+    {									\
+      if (SCM_UNLIKELY (scm_is_false (srfi1_module)))			\
+	srfi1_module = scm_c_resolve_module ("srfi srfi-1");		\
+									\
+      var = scm_module_variable (srfi1_module,				\
+                                 scm_from_locale_symbol (name));	\
+      if (SCM_UNLIKELY (scm_is_false (var)))				\
+        abort ();							\
+									\
+      var = SCM_VARIABLE_REF (var);					\
+    }
+
 static long
 srfi1_ilength (SCM sx)
 {
@@ -186,83 +207,26 @@ SCM_DEFINE (scm_srfi1_append_reverse_x, "append-reverse!", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_break, "break", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return two values, the longest initial prefix of @var{lst}\n"
-	    "whose elements all fail the predicate @var{pred}, and the\n"
-	    "remainder of @var{lst}.\n"
-	    "\n"
-	    "Note that the name @code{break} conflicts with the @code{break}\n"
-	    "binding established by @code{while}.  Applications wanting to\n"
-	    "use @code{break} from within a @code{while} loop will need to\n"
-	    "make a new define under a different name.")
-#define FUNC_NAME s_scm_srfi1_break
+SCM
+scm_srfi1_break (SCM pred, SCM lst)
 {
-  SCM ret, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  ret = SCM_EOL;
-  p = &ret;
-  for ( ; scm_is_pair (lst); lst = SCM_CDR (lst))
-    {
-      SCM elem = SCM_CAR (lst);
-      if (scm_is_true (scm_call_1 (pred, elem)))
-        goto done;
-
-      /* want this elem, tack it onto the end of ret */
-      *p = scm_cons (elem, SCM_EOL);
-      p = SCM_CDRLOC (*p);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  return scm_values (scm_list_2 (ret, lst));
+  CACHE_VAR (break_proc, "break");
+  return scm_call_2 (break_proc, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_break_x, "break!", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return two values, the longest initial prefix of @var{lst}\n"
-	    "whose elements all fail the predicate @var{pred}, and the\n"
-	    "remainder of @var{lst}.  @var{lst} may be modified to form the\n"
-	    "return.")
-#define FUNC_NAME s_scm_srfi1_break_x
+SCM
+scm_srfi1_break_x (SCM pred, SCM lst)
 {
-  SCM upto, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  p = &lst;
-  for (upto = lst; scm_is_pair (upto); upto = SCM_CDR (upto))
-    {
-      if (scm_is_true (scm_call_1 (pred, SCM_CAR (upto))))
-        goto done;
-
-      /* want this element */
-      p = SCM_CDRLOC (upto);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (upto), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  *p = SCM_EOL;
-  return scm_values (scm_list_2 (lst, upto));
+  CACHE_VAR (break_x, "break!");
+  return scm_call_2 (break_x, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_car_plus_cdr, "car+cdr", 1, 0, 0,
-            (SCM pair),
-	    "Return two values, the @sc{car} and the @sc{cdr} of @var{pair}.")
-#define FUNC_NAME s_scm_srfi1_car_plus_cdr
+SCM
+scm_srfi1_car_plus_cdr (SCM pair)
 {
-  SCM_VALIDATE_CONS (SCM_ARG1, pair);
-  return scm_values (scm_list_2 (SCM_CAR (pair), SCM_CDR (pair)));
+  CACHE_VAR (car_plus_cdr, "car+cdr");
+  return scm_call_1 (car_plus_cdr, pair);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_concatenate, "concatenate", 1, 0, 0,
             (SCM lstlst),
@@ -719,71 +683,33 @@ SCM_DEFINE (scm_srfi1_drop_right, "drop-right", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_drop_right_x, "drop-right!", 2, 0, 0,
-            (SCM lst, SCM n),
-	    "Return the a list containing the @var{n} last elements of\n"
-	    "@var{lst}.  @var{lst} may be modified to build the return.")
-#define FUNC_NAME s_scm_srfi1_drop_right_x
+SCM
+scm_srfi1_drop_right_x (SCM lst, SCM n)
 {
-  SCM tail, *p;
-
-  if (scm_is_eq (n, SCM_INUM0))
-    return lst;
-
-  tail = scm_list_tail (lst, n);
-  p = &lst;
-
-  /* p and tail work along the list, p being the cdrloc of the cell n steps
-     behind tail */
-  for ( ; scm_is_pair (tail); tail = SCM_CDR (tail))
-    p = SCM_CDRLOC (*p);
-
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P(tail), tail, SCM_ARG1, FUNC_NAME, "list");
-
-  *p = SCM_EOL;
-  return lst;
+  CACHE_VAR (drop_right_x, "drop-right!");
+  return scm_call_2 (drop_right_x, lst, n);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_drop_while, "drop-while", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Drop the longest initial prefix of @var{lst} whose elements all\n"
-	    "satisfy the predicate @var{pred}.")
-#define FUNC_NAME s_scm_srfi1_drop_while
+SCM
+scm_srfi1_drop_while (SCM pred, SCM lst)
 {
-  SCM_VALIDATE_PROC (SCM_ARG1, pred);
-
-  for ( ; scm_is_pair (lst); lst = SCM_CDR (lst))
-    if (scm_is_false (scm_call_1 (pred, SCM_CAR (lst))))
-      goto done;
-
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, SCM_ARG2, FUNC_NAME, "list");
- done:
-  return lst;
+  CACHE_VAR (drop_while, "drop-while");
+  return scm_call_2 (drop_while, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_eighth, "eighth", 1, 0, 0,
-            (SCM lst),
-	    "Return the eighth element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_eighth
+SCM
+scm_srfi1_eighth (SCM lst)
 {
-  return scm_list_ref (lst, SCM_I_MAKINUM (7));
+  CACHE_VAR (eighth, "eighth");
+  return scm_call_1 (eighth, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_fifth, "fifth", 1, 0, 0,
-            (SCM lst),
-	    "Return the fifth element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_fifth
+SCM
+scm_srfi1_fifth (SCM lst)
 {
-  return scm_list_ref (lst, SCM_I_MAKINUM (4));
+  CACHE_VAR (fifth, "fifth");
+  return scm_call_1 (fifth, lst);
 }
-#undef FUNC_NAME
 
 
 SCM_DEFINE (scm_srfi1_filter_map, "filter-map", 2, 0, 1,
@@ -935,131 +861,19 @@ SCM_DEFINE (scm_srfi1_find_tail, "find-tail", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_fold, "fold", 3, 0, 1,
-            (SCM proc, SCM init, SCM list1, SCM rest),
-	    "Apply @var{proc} to the elements of @var{lst1} @dots{}\n"
-	    "@var{lstN} to build a result, and return that result.\n"
-	    "\n"
-	    "Each @var{proc} call is @code{(@var{proc} @var{elem1} @dots{}\n"
-	    "@var{elemN} @var{previous})}, where @var{elem1} is from\n"
-	    "@var{lst1}, through @var{elemN} from @var{lstN}.\n"
-	    "@var{previous} is the return from the previous call to\n"
-	    "@var{proc}, or the given @var{init} for the first call.  If any\n"
-	    "list is empty, just @var{init} is returned.\n"
-	    "\n"
-	    "@code{fold} works through the list elements from first to last.\n"
-	    "The following shows a list reversal and the calls it makes,\n"
-	    "\n"
-	    "@example\n"
-	    "(fold cons '() '(1 2 3))\n"
-	    "\n"
-	    "(cons 1 '())\n"
-	    "(cons 2 '(1))\n"
-	    "(cons 3 '(2 1)\n"
-	    "@result{} (3 2 1)\n"
-	    "@end example\n"
-	    "\n"
-	    "If @var{lst1} through @var{lstN} have different lengths,\n"
-	    "@code{fold} stops when the end of the shortest is reached.\n"
-	    "Ie.@: elements past the length of the shortest are ignored in\n"
-	    "the other @var{lst}s.  At least one @var{lst} must be\n"
-	    "non-circular.\n"
-	    "\n"
-	    "The way @code{fold} builds a result from iterating is quite\n"
-	    "general, it can do more than other iterations like say\n"
-	    "@code{map} or @code{filter}.  The following for example removes\n"
-	    "adjacent duplicate elements from a list,\n"
-	    "\n"
-	    "@example\n"
-	    "(define (delete-adjacent-duplicates lst)\n"
-	    "  (fold-right (lambda (elem ret)\n"
-	    "                (if (equal? elem (first ret))\n"
-	    "                    ret\n"
-	    "                    (cons elem ret)))\n"
-	    "              (list (last lst))\n"
-	    "              lst))\n"
-	    "(delete-adjacent-duplicates '(1 2 3 3 4 4 4 5))\n"
-	    "@result{} (1 2 3 4 5)\n"
-	    "@end example\n"
-	    "\n"
-	    "Clearly the same sort of thing can be done with a\n"
-	    "@code{for-each} and a variable in which to build the result,\n"
-	    "but a self-contained @var{proc} can be re-used in multiple\n"
-	    "contexts, where a @code{for-each} would have to be written out\n"
-	    "each time.")
-#define FUNC_NAME s_scm_srfi1_fold
+SCM
+scm_srfi1_fold (SCM proc, SCM init, SCM list1, SCM rest)
 {
-  SCM lst;
-  int argnum;
-  SCM_VALIDATE_REST_ARGUMENT (rest);
-
-  if (scm_is_null (rest))
-    {
-      /* one list */
-      SCM_VALIDATE_PROC (SCM_ARG1, proc);
-
-      for ( ; scm_is_pair (list1); list1 = SCM_CDR (list1))
-        init = scm_call_2 (proc, SCM_CAR (list1), init);
-
-      /* check below that list1 is a proper list, and done */
-      lst = list1;
-      argnum = 2;
-    }
-  else
-    {
-      /* two or more lists */
-      SCM  vec, args, a;
-      size_t  len, i;
-
-      /* vec is the list arguments */
-      vec = scm_vector (scm_cons (list1, rest));
-      len = SCM_SIMPLE_VECTOR_LENGTH (vec);
-
-      /* args is the argument list to pass to proc, same length as vec,
-         re-used for each call */
-      args = scm_make_list (SCM_I_MAKINUM (len+1), SCM_UNDEFINED);
-
-      for (;;)
-        {
-          /* first elem of each list in vec into args, and step those
-             vec entries onto their next element */
-          for (i = 0, a = args, argnum = 2;
-               i < len;
-               i++, a = SCM_CDR (a), argnum++)
-            {
-              lst = SCM_SIMPLE_VECTOR_REF (vec, i);  /* list argument */
-              if (! scm_is_pair (lst))
-                goto check_lst_and_done;
-              SCM_SETCAR (a, SCM_CAR (lst));  /* arg for proc */
-              SCM_SIMPLE_VECTOR_SET (vec, i, SCM_CDR (lst));  /* rest of lst */
-            }
-          SCM_SETCAR (a, init);
-
-          init = scm_apply (proc, args, SCM_EOL);
-        }
-    }
-
- check_lst_and_done:
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, argnum, FUNC_NAME, "list");
-  return init;
+  CACHE_VAR (fold, "fold");
+  return scm_apply_3 (fold, proc, init, list1, rest);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_last, "last", 1, 0, 0,
-            (SCM lst),
-	    "Like @code{cons}, but with interchanged arguments.  Useful\n"
-	    "mostly when passed to higher-order procedures.")
-#define FUNC_NAME s_scm_srfi1_last
+SCM
+scm_srfi1_last (SCM lst)
 {
-  SCM pair = scm_last_pair (lst);
-  /* scm_last_pair returns SCM_EOL for an empty list */
-  SCM_VALIDATE_CONS (SCM_ARG1, pair);
-  return SCM_CAR (pair);
+  CACHE_VAR (last, "last");
+  return scm_call_1 (last, lst);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_length_plus, "length+", 1, 0, 0,
             (SCM lst),
@@ -1073,106 +887,12 @@ SCM_DEFINE (scm_srfi1_length_plus, "length+", 1, 0, 0,
 #undef FUNC_NAME
 
 
-SCM_DEFINE (scm_srfi1_list_index, "list-index", 2, 0, 1,
-            (SCM pred, SCM list1, SCM rest),
-	    "Return the index of the first set of elements, one from each of\n"
-	    "@var{lst1}@dots{}@var{lstN}, which satisfies @var{pred}.\n"
-	    "\n"
-	    "@var{pred} is called as @code{(@var{pred} elem1 @dots{}\n"
-	    "elemN)}.  Searching stops when the end of the shortest\n"
-	    "@var{lst} is reached.  The return index starts from 0 for the\n"
-	    "first set of elements.  If no set of elements pass then the\n"
-	    "return is @code{#f}.\n"
-	    "\n"
-	    "@example\n"
-	    "(list-index odd? '(2 4 6 9))      @result{} 3\n"
-	    "(list-index = '(1 2 3) '(3 1 2))  @result{} #f\n"
-	    "@end example")
-#define FUNC_NAME s_scm_srfi1_list_index
+SCM
+scm_srfi1_list_index (SCM pred, SCM list1, SCM rest)
 {
-  long  n = 0;
-  SCM   lst;
-  int   argnum;
-  SCM_VALIDATE_REST_ARGUMENT (rest);
-
-  if (scm_is_null (rest))
-    {
-      /* one list */
-      SCM_VALIDATE_PROC (SCM_ARG1, pred);
-
-      for ( ; scm_is_pair (list1); n++, list1 = SCM_CDR (list1))
-        if (scm_is_true (scm_call_1 (pred, SCM_CAR (list1))))
-          return SCM_I_MAKINUM (n);
-
-      /* not found, check below that list1 is a proper list */
-    end_list1:
-      lst = list1;
-      argnum = 2;
-    }
-  else if (scm_is_pair (rest) && scm_is_null (SCM_CDR (rest)))
-    {
-      /* two lists */
-      SCM list2 = SCM_CAR (rest);
-      SCM_VALIDATE_PROC (SCM_ARG1, pred);
-
-      for ( ; ; n++)
-        {
-          if (! scm_is_pair (list1))
-            goto end_list1;
-          if (! scm_is_pair (list2))
-            {
-              lst = list2;
-              argnum = 3;
-              break;
-            }
-          if (scm_is_true (scm_call_2 (pred,
-                                       SCM_CAR (list1), SCM_CAR (list2))))
-            return SCM_I_MAKINUM (n);
-
-          list1 = SCM_CDR (list1);
-          list2 = SCM_CDR (list2);
-        }
-    }
-  else
-    {
-      /* three or more lists */
-      SCM     vec, args, a;
-      size_t  len, i;
-
-      /* vec is the list arguments */
-      vec = scm_vector (scm_cons (list1, rest));
-      len = SCM_SIMPLE_VECTOR_LENGTH (vec);
-
-      /* args is the argument list to pass to pred, same length as vec,
-         re-used for each call */
-      args = scm_make_list (SCM_I_MAKINUM (len), SCM_UNDEFINED);
-
-      for ( ; ; n++)
-        {
-          /* first elem of each list in vec into args, and step those
-             vec entries onto their next element */
-          for (i = 0, a = args, argnum = 2;
-               i < len;
-               i++, a = SCM_CDR (a), argnum++)
-            {
-              lst = SCM_SIMPLE_VECTOR_REF (vec, i);  /* list argument */
-              if (! scm_is_pair (lst))
-                goto not_found_check_lst;
-              SCM_SETCAR (a, SCM_CAR (lst));  /* arg for pred */
-              SCM_SIMPLE_VECTOR_SET (vec, i, SCM_CDR (lst));  /* rest of lst */
-            }
-
-          if (scm_is_true (scm_apply (pred, args, SCM_EOL)))
-            return SCM_I_MAKINUM (n);
-        }
-    }
-
- not_found_check_lst:
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, argnum, FUNC_NAME, "list");
-  return SCM_BOOL_F;
+  CACHE_VAR (list_index, "list-index");
+  return scm_apply_2 (list_index, pred, list1, rest);
 }
-#undef FUNC_NAME
-
 
 /* This routine differs from the core list-copy in allowing improper lists.
    Maybe the core could allow them similarly.  */
@@ -1206,75 +926,19 @@ SCM_DEFINE (scm_srfi1_list_copy, "list-copy", 1, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_list_tabulate, "list-tabulate", 2, 0, 0,
-            (SCM n, SCM proc),
-	    "Return an @var{n}-element list, where each list element is\n"
-	    "produced by applying the procedure @var{init-proc} to the\n"
-	    "corresponding list index.  The order in which @var{init-proc}\n"
-	    "is applied to the indices is not specified.")
-#define FUNC_NAME s_scm_srfi1_list_tabulate
+SCM
+scm_srfi1_list_tabulate (SCM n, SCM proc)
 {
-  long i, nn;
-  SCM ret = SCM_EOL;
-  nn = scm_to_signed_integer (n, 0, LONG_MAX);
-  SCM_VALIDATE_PROC (SCM_ARG2, proc);
-  for (i = nn-1; i >= 0; i--)
-    ret = scm_cons (scm_call_1 (proc, scm_from_long (i)), ret);
-  return ret;
+  CACHE_VAR (list_tabulate, "list-tabulate");
+  return scm_call_2 (list_tabulate, n, proc);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_lset_adjoin, "lset-adjoin", 2, 0, 1,
-            (SCM equal, SCM lst, SCM rest),
-	    "Add to @var{list} any of the given @var{elem}s not already in\n"
-	    "the list.  @var{elem}s are @code{cons}ed onto the start of\n"
-	    "@var{list} (so the return shares a common tail with\n"
-	    "@var{list}), but the order they're added is unspecified.\n"
-	    "\n"
-	    "The given @var{=} procedure is used for comparing elements,\n"
-	    "called as @code{(@var{=} listelem elem)}, ie.@: the second\n"
-	    "argument is one of the given @var{elem} parameters.\n"
-	    "\n"
-	    "@example\n"
-	    "(lset-adjoin eqv? '(1 2 3) 4 1 5) @result{} (5 4 1 2 3)\n"
-	    "@end example")
-#define FUNC_NAME s_scm_srfi1_lset_adjoin
+SCM
+scm_srfi1_lset_adjoin (SCM equal, SCM lst, SCM rest)
 {
-  SCM l, elem;
-
-  SCM_VALIDATE_PROC (SCM_ARG1, equal);
-  SCM_VALIDATE_REST_ARGUMENT (rest);
-
-  /* It's not clear if duplicates among the `rest' elements are meant to be
-     cast out.  The spec says `=' is called as (= list-elem rest-elem),
-     suggesting perhaps not, but the reference implementation shows the
-     "list" at each stage as including those "rest" elements already added.
-     The latter corresponds to what's described for lset-union, so that's
-     what's done here.  */
-
-  for ( ; scm_is_pair (rest); rest = SCM_CDR (rest))
-    {
-      elem = SCM_CAR (rest);
-
-      for (l = lst; scm_is_pair (l); l = SCM_CDR (l))
-        if (scm_is_true (scm_call_2 (equal, SCM_CAR (l), elem)))
-          goto next_elem; /* elem already in lst, don't add */
-
-      SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P(l), lst, SCM_ARG2, FUNC_NAME, "list");
-
-      /* elem is not equal to anything already in lst, add it */
-      lst = scm_cons (elem, lst);
-
-    next_elem:
-      ;
-    }
-
-  return lst;
+  CACHE_VAR (lset_adjoin, "lset-adjoin");
+  return scm_apply_1 (lset_adjoin, lst, rest);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_lset_difference_x, "lset-difference!", 2, 0, 1,
             (SCM equal, SCM lst, SCM rest),
@@ -1600,30 +1264,19 @@ SCM_DEFINE (scm_srfi1_assoc, "assoc", 2, 1, 0,
 #undef FUNC_NAME
 
 
-SCM_DEFINE (scm_srfi1_ninth, "ninth", 1, 0, 0,
-            (SCM lst),
-	    "Return the ninth element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_ninth
+SCM
+scm_srfi1_ninth (SCM lst)
 {
-  return scm_list_ref (lst, scm_from_int (8));
+  CACHE_VAR (ninth, "ninth");
+  return scm_call_1 (ninth, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_not_pair_p, "not-pair?", 1, 0, 0,
-            (SCM obj),
-	    "Return @code{#t} is @var{obj} is not a pair, @code{#f}\n"
-	    "otherwise.\n"
-	    "\n"
-	    "This is shorthand notation @code{(not (pair?  @var{obj}))} and\n"
-	    "is supposed to be used for end-of-list checking in contexts\n"
-	    "where dotted lists are allowed.")
-#define FUNC_NAME s_scm_srfi1_not_pair_p
+SCM
+scm_srfi1_not_pair_p (SCM obj)
 {
-  return scm_from_bool (! scm_is_pair (obj));
+  CACHE_VAR (not_pair_p, "not-pair?");
+  return scm_call_1 (not_pair_p, obj);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_partition, "partition", 2, 0, 0,
 	    (SCM pred, SCM list),
@@ -1725,126 +1378,19 @@ SCM_DEFINE (scm_srfi1_partition_x, "partition!", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_reduce, "reduce", 3, 0, 0,
-            (SCM proc, SCM def, SCM lst),
-	    "@code{reduce} is a variant of @code{fold}, where the first call\n"
-	    "to @var{proc} is on two elements from @var{lst}, rather than\n"
-	    "one element and a given initial value.\n"
-	    "\n"
-	    "If @var{lst} is empty, @code{reduce} returns @var{def} (this is\n"
-	    "the only use for @var{def}).  If @var{lst} has just one element\n"
-	    "then that's the return value.  Otherwise @var{proc} is called\n"
-	    "on the elements of @var{lst}.\n"
-	    "\n"
-	    "Each @var{proc} call is @code{(@var{proc} @var{elem}\n"
-	    "@var{previous})}, where @var{elem} is from @var{lst} (the\n"
-	    "second and subsequent elements of @var{lst}), and\n"
-	    "@var{previous} is the return from the previous call to\n"
-	    "@var{proc}.  The first element of @var{lst} is the\n"
-	    "@var{previous} for the first call to @var{proc}.\n"
-	    "\n"
-	    "For example, the following adds a list of numbers, the calls\n"
-	    "made to @code{+} are shown.  (Of course @code{+} accepts\n"
-	    "multiple arguments and can add a list directly, with\n"
-	    "@code{apply}.)\n"
-	    "\n"
-	    "@example\n"
-	    "(reduce + 0 '(5 6 7)) @result{} 18\n"
-	    "\n"
-	    "(+ 6 5)  @result{} 11\n"
-	    "(+ 7 11) @result{} 18\n"
-	    "@end example\n"
-	    "\n"
-	    "@code{reduce} can be used instead of @code{fold} where the\n"
-	    "@var{init} value is an ``identity'', meaning a value which\n"
-	    "under @var{proc} doesn't change the result, in this case 0 is\n"
-	    "an identity since @code{(+ 5 0)} is just 5.  @code{reduce}\n"
-	    "avoids that unnecessary call.")
-#define FUNC_NAME s_scm_srfi1_reduce
+SCM
+scm_srfi1_reduce (SCM proc, SCM def, SCM lst)
 {
-  SCM ret;
-  SCM_VALIDATE_PROC (SCM_ARG1, proc);
-  ret = def;  /* if lst is empty */
-  if (scm_is_pair (lst))
-    {
-      ret = SCM_CAR (lst);  /* if lst has one element */
-
-      for (lst = SCM_CDR (lst); scm_is_pair (lst); lst = SCM_CDR (lst))
-        ret = scm_call_2 (proc, SCM_CAR (lst), ret);
-    }
-
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, SCM_ARG3, FUNC_NAME, "list");
-  return ret;
+  CACHE_VAR (reduce, "reduce");
+  return scm_call_3 (reduce, proc, def, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_reduce_right, "reduce-right", 3, 0, 0,
-            (SCM proc, SCM def, SCM lst),
-	    "@code{reduce-right} is a variant of @code{fold-right}, where\n"
-	    "the first call to @var{proc} is on two elements from @var{lst},\n"
-	    "rather than one element and a given initial value.\n"
-	    "\n"
-	    "If @var{lst} is empty, @code{reduce-right} returns @var{def}\n"
-	    "(this is the only use for @var{def}).  If @var{lst} has just\n"
-	    "one element then that's the return value.  Otherwise @var{proc}\n"
-	    "is called on the elements of @var{lst}.\n"
-	    "\n"
-	    "Each @var{proc} call is @code{(@var{proc} @var{elem}\n"
-	    "@var{previous})}, where @var{elem} is from @var{lst} (the\n"
-	    "second last and then working back to the first element of\n"
-	    "@var{lst}), and @var{previous} is the return from the previous\n"
-	    "call to @var{proc}.  The last element of @var{lst} is the\n"
-	    "@var{previous} for the first call to @var{proc}.\n"
-	    "\n"
-	    "For example, the following adds a list of numbers, the calls\n"
-	    "made to @code{+} are shown.  (Of course @code{+} accepts\n"
-	    "multiple arguments and can add a list directly, with\n"
-	    "@code{apply}.)\n"
-	    "\n"
-	    "@example\n"
-	    "(reduce-right + 0 '(5 6 7)) @result{} 18\n"
-	    "\n"
-	    "(+ 6 7)  @result{} 13\n"
-	    "(+ 5 13) @result{} 18\n"
-	    "@end example\n"
-	    "\n"
-	    "@code{reduce-right} can be used instead of @code{fold-right}\n"
-	    "where the @var{init} value is an ``identity'', meaning a value\n"
-	    "which under @var{proc} doesn't change the result, in this case\n"
-	    "0 is an identity since @code{(+ 7 0)} is just 5.\n"
-	    "@code{reduce-right} avoids that unnecessary call.\n"
-	    "\n"
-	    "@code{reduce} should be preferred over @code{reduce-right} if\n"
-	    "the order of processing doesn't matter, or can be arranged\n"
-	    "either way, since @code{reduce} is a little more efficient.")
-#define FUNC_NAME s_scm_srfi1_reduce_right
+SCM
+scm_srfi1_reduce_right (SCM proc, SCM def, SCM lst)
 {
-  /* To work backwards across a list requires either repeatedly traversing
-     to get each previous element, or using some memory for a reversed or
-     random-access form.  Repeated traversal might not be too terrible, but
-     is of course quadratic complexity and hence to be avoided in case LST
-     is long.  A vector is preferred over a reversed list since it's more
-     compact and is less work for the gc to collect.  */
-
-  SCM vec, ret;
-  ssize_t len, i;
-  SCM_VALIDATE_PROC (SCM_ARG1, proc);
-  if (SCM_NULL_OR_NIL_P (lst))
-    return def;
-
-  vec = scm_vector (lst);
-  len = SCM_SIMPLE_VECTOR_LENGTH (vec);
-
-  ret = SCM_SIMPLE_VECTOR_REF (vec, len-1);
-  for (i = len-2; i >= 0; i--)
-    ret = scm_call_2 (proc, SCM_SIMPLE_VECTOR_REF (vec, i), ret);
-
-  return ret;
+  CACHE_VAR (reduce_right, "reduce-right");
+  return scm_call_3 (reduce_right, proc, def, lst);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_remove, "remove", 2, 0, 0,
 	    (SCM pred, SCM list),
@@ -1907,86 +1453,33 @@ SCM_DEFINE (scm_srfi1_remove_x, "remove!", 2, 0, 0,
 #undef FUNC_NAME
 
 
-SCM_DEFINE (scm_srfi1_seventh, "seventh", 1, 0, 0,
-            (SCM lst),
-	    "Return the seventh element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_seventh
+SCM
+scm_srfi1_seventh (SCM lst)
 {
-  return scm_list_ref (lst, scm_from_int (6));
+  CACHE_VAR (seventh, "seventh");
+  return scm_call_1 (seventh, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_sixth, "sixth", 1, 0, 0,
-            (SCM lst),
-	    "Return the sixth element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_sixth
+SCM
+scm_srfi1_sixth (SCM lst)
 {
-  return scm_list_ref (lst, scm_from_int (5));
+  CACHE_VAR (sixth, "sixth");
+  return scm_call_1 (sixth, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_span, "span", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return two values, the longest initial prefix of @var{lst}\n"
-	    "whose elements all satisfy the predicate @var{pred}, and the\n"
-	    "remainder of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_span
+SCM
+scm_srfi1_span (SCM pred, SCM lst)
 {
-  SCM ret, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  ret = SCM_EOL;
-  p = &ret;
-  for ( ; scm_is_pair (lst); lst = SCM_CDR (lst))
-    {
-      SCM elem = SCM_CAR (lst);
-      if (scm_is_false (scm_call_1 (pred, elem)))
-        goto done;
-
-      /* want this elem, tack it onto the end of ret */
-      *p = scm_cons (elem, SCM_EOL);
-      p = SCM_CDRLOC (*p);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  return scm_values (scm_list_2 (ret, lst));
+  CACHE_VAR (span, "span");
+  return scm_call_2 (span, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_span_x, "span!", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return two values, the longest initial prefix of @var{lst}\n"
-	    "whose elements all satisfy the predicate @var{pred}, and the\n"
-	    "remainder of @var{lst}.  @var{lst} may be modified to form the\n"
-	    "return.")
-#define FUNC_NAME s_scm_srfi1_span_x
+SCM
+scm_srfi1_span_x (SCM pred, SCM lst)
 {
-  SCM upto, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  p = &lst;
-  for (upto = lst; scm_is_pair (upto); upto = SCM_CDR (upto))
-    {
-      if (scm_is_false (scm_call_1 (pred, SCM_CAR (upto))))
-        goto done;
-
-      /* want this element */
-      p = SCM_CDRLOC (upto);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (upto), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  *p = SCM_EOL;
-  return scm_values (scm_list_2 (lst, upto));
+  CACHE_VAR (span_x, "span!");
+  return scm_call_2 (span_x, pred, lst);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_split_at, "split-at", 2, 0, 0,
             (SCM lst, SCM n),
@@ -2038,33 +1531,12 @@ SCM_DEFINE (scm_srfi1_split_at_x, "split-at!", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_take_x, "take!", 2, 0, 0,
-            (SCM lst, SCM n),
-	    "Return a list containing the first @var{n} elements of\n"
-	    "@var{lst}.")
-#define FUNC_NAME s_scm_srfi1_take_x
+SCM
+scm_srfi1_take_x (SCM lst, SCM n)
 {
-  long nn;
-  SCM pos;
-
-  nn = scm_to_signed_integer (n, 0, LONG_MAX);
-  if (nn == 0)
-    return SCM_EOL;
-
-  pos = scm_list_tail (lst, scm_from_long (nn - 1));
-
-  /* Must have at least one cell left, mustn't have reached the end of an
-     n-1 element list.  SCM_VALIDATE_CONS here gives the same error as
-     scm_list_tail does on say an n-2 element list, though perhaps a range
-     error would make more sense (for both).  */
-  SCM_VALIDATE_CONS (SCM_ARG1, pos);
-
-  SCM_SETCDR (pos, SCM_EOL);
-  return lst;
+  CACHE_VAR (take_x, "take!");
+  return scm_call_2 (take_x, lst, n);
 }
-#undef FUNC_NAME
-
 
 SCM_DEFINE (scm_srfi1_take_right, "take-right", 2, 0, 0,
             (SCM lst, SCM n),
@@ -2084,86 +1556,35 @@ SCM_DEFINE (scm_srfi1_take_right, "take-right", 2, 0, 0,
 #undef FUNC_NAME
 
 
-SCM_DEFINE (scm_srfi1_take_while, "take-while", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return a new list which is the longest initial prefix of\n"
-	    "@var{lst} whose elements all satisfy the predicate @var{pred}.")
-#define FUNC_NAME s_scm_srfi1_take_while
+SCM
+scm_srfi1_take_while (SCM pred, SCM lst)
 {
-  SCM ret, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  ret = SCM_EOL;
-  p = &ret;
-  for ( ; scm_is_pair (lst); lst = SCM_CDR (lst))
-    {
-      SCM elem = SCM_CAR (lst);
-      if (scm_is_false (scm_call_1 (pred, elem)))
-        goto done;
-
-      /* want this elem, tack it onto the end of ret */
-      *p = scm_cons (elem, SCM_EOL);
-      p = SCM_CDRLOC (*p);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (lst), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  return ret;
+  CACHE_VAR (take_while, "take-while");
+  return scm_call_2 (take_while, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_take_while_x, "take-while!", 2, 0, 0,
-            (SCM pred, SCM lst),
-	    "Return the longest initial prefix of @var{lst} whose elements\n"
-	    "all satisfy the predicate @var{pred}.  @var{lst} may be\n"
-	    "modified to form the return.")
-#define FUNC_NAME s_scm_srfi1_take_while_x
+SCM
+scm_srfi1_take_while_x (SCM pred, SCM lst)
 {
-  SCM upto, *p;
-
-  SCM_ASSERT (scm_is_true (scm_procedure_p (pred)), pred, SCM_ARG1, FUNC_NAME);
-
-  p = &lst;
-  for (upto = lst; scm_is_pair (upto); upto = SCM_CDR (upto))
-    {
-      if (scm_is_false (scm_call_1 (pred, SCM_CAR (upto))))
-        goto done;
-
-      /* want this element */
-      p = SCM_CDRLOC (upto);
-    }
-  SCM_ASSERT_TYPE (SCM_NULL_OR_NIL_P (upto), lst, SCM_ARG2, FUNC_NAME, "list");
-
- done:
-  *p = SCM_EOL;
-  return lst;
+  CACHE_VAR (take_while_x, "take-while!");
+  return scm_call_2 (take_while_x, pred, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_tenth, "tenth", 1, 0, 0,
-            (SCM lst),
-	    "Return the tenth element of @var{lst}.")
-#define FUNC_NAME s_scm_srfi1_tenth
+SCM
+scm_srfi1_tenth (SCM lst)
 {
-  return scm_list_ref (lst, scm_from_int (9));
+  CACHE_VAR (tenth, "tenth");
+  return scm_call_1 (tenth, lst);
 }
-#undef FUNC_NAME
 
-
-SCM_DEFINE (scm_srfi1_xcons, "xcons", 2, 0, 0,
-            (SCM d, SCM a),
-	    "Like @code{cons}, but with interchanged arguments.  Useful\n"
-	    "mostly when passed to higher-order procedures.")
-#define FUNC_NAME s_scm_srfi1_xcons
+SCM
+scm_srfi1_xcons (SCM d, SCM a)
 {
-  return scm_cons (a, d);
+  CACHE_VAR (xcons, "xcons");
+  return scm_call_2 (xcons, d, a);
 }
-#undef FUNC_NAME
-
 
+
 void
 scm_init_srfi_1 (void)
 {
